/* File: PCR.h - last edit:
 * MXT		Thu Nov 21 15:04:59 1991
 */

/* Copyright (C) 1992 by Digital Equipment Corporation and Xerox Corporation */

/*	All Rights Reserved
Permission to use, copy, modify, and distribute this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Digital and Xerox not
be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  Digital and
Xerox make no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.
*/

/*
 * This is to make xlib work in conjunction with the multi-threaded environment
 * of PCR
 */

#ifndef _PCR_H
#define _PCR_H

/*
 * The following header files will cause errors if they are included after
 * the PCR header files that have the systems calls #define'd
 */
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <netdb.h>
#include <signal.h>
#include <pwd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <X11/Xosdefs.h>

#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#include <string.h>
#endif

/* PCR include files defining PCR versions of UNIX system calls */
#include	"xr/Threads.h"
#include	"xr/ThreadsMsg.h"
#include	"xr/UIO.h"
#include	"xr/Errno.h"

/* use PCR's memory allocation functions */
extern char *XR_malloc();
extern char *XR_calloc();
extern char *XR_realloc();
extern void XR_free();

#define Xmalloc(size) XR_malloc((size))
#define Xcalloc(nelem, elsize) XR_calloc((nelem), (elsize))
#define Xrealloc(ptr, size) XR_realloc((ptr), (size))
#ifndef DEBUG_BUFF
#define Xfree(ptr) XR_free((ptr))
#endif

/* functions returning pointers which need to be prototyped */
struct hostent *XR_GetHostByName();
struct passwd *XR_GetPWNam();
struct passwd *XR_GetPWUID();

/* these are system routines that have equivalent counterparts in PCR */
#define	access		XR_Access
#define	close		XR_Close
#define	connect		XR_Connect
#define	exit(a)		XR_FPrintF(XR_MSG_STDOUT, "Xlib calls exit(%d).\n",(a));\
			XR_CallDebugger()
#define fcntl		XR_FCntl
#define	getenv		XR_getenv
#define gethostname	XR_GetHostName
#define gethostbyname	xr_gethostbyname
#define getpwnam	XR_GetPWNam
#define getpwuid	XR_GetPWUID
#define gettimeofday	XR_GetTimeOfDay
#define	getuid		XR_GetUID
#define ioctl		XR_IOCtl
#define open		XR_Open
#define qsort           XR_qsort
#define	read		XR_Read
#define	readv		XR_ReadV
#define rewind		XR_rewind
#define	setsockopt	XR_SetSockOpt
#define	shmdt		XR_ShmDt
#define	shmctl		XR_ShmCtl
#define	shmat		XR_ShmAt
#define	shmget		XR_ShmGet
#define	socket		XR_Socket
#define	sprintf		XR_SPrintF
#define sscanf		XR_sscanf
#define	write		XR_Write
#define	writev		XR_WriteV

#ifdef OSF1
#undef	readv
#undef	writev
#define	readv		XR_XReadV
#define	writev		XR_XWriteV
#endif

/* sleep is translated to XR_PauseAbortable and seconds translated to ticks */
#define	sleep(secs)	XR_PauseAbortable( (secs) * XR_TICKS_PER_SECOND )

/* errno is redefined in PCR to be a per-thread value */
#define	errno		XR_errno

#ifdef NOTUSED	/* moved to Threads.h */

/* thread locking, waiting, notifying, and more */

#define	Mutex			struct XR_MLRep
#define	NullMutex		XR_NULL_ML_REP

/* Mutex *mtx */
#define	InitMutex(mtx)		XR_InitializeMonitor((mtx))
#define DestroyMutex(mtx)
#define LockMutex(mtx)		XR_MonitorEntry((mtx))
#define UnlockMutex(mtx)	XR_MonitorExit((mtx))

#define	Condition		struct XR_CVRep
#define	NullCondition		XR_NULL_Condition_REP

/* Condition *cond, Mutex *mtx, Timeout timeout */
#define	InitCondition(cond, timeout)	XR_InitializeCondition((cond), (timeout))
#define DestroyCondition(cond)
#define WaitCondition(cond, mtx, timeout)	/* timeout in milliseconds */ \
	if ((timeout) != NOWAIT) {\
	    if ((timeout) != WAITFOREVER)\
		XR_SetTimeout((cond), ((timeout)+XR_MSECS_PER_TICK-1)/XR_MSECS_PER_TICK);\
	    XR_WaitCV((cond), (mtx));\
	    if (XR_GetTimeout((cond)) != XR_WAIT_FOREVER)\
		XR_DisableTimeout((cond));\
	};
#define	NotifyCondition(cond)		XR_Notify((cond))
#define	BroadcastCondition(cond)	XR_Broadcast((cond))

/* Display *dpy */
#define LockDisplay(dpy)	LockMutex(&(dpy)->locks->general)
#define UnlockDisplay(dpy)	UnlockMutex(&(dpy)->locks->general)

#endif /* NOTUSED */

int XR_XReadV(
#if NeedFunctionPrototypes
    XR_Fildes,		/* fildes */
    struct iovec *,	/* iov */
    int			/* iovCnt */
#endif
);

int XR_XWriteV(
#if NeedFunctionPrototypes
    XR_Fildes,		/* fildes */
    struct iovec *,	/* iov */
    int			/* iovCnt */
#endif
);

/*
   A truly simple thread-safe stdio.  There is NO support for seeking, or for
    switching between read and write on a particular file pointer.
*/

#ifndef BUFSIZ
#define BUFSIZ	512
#endif

#ifndef EOF
#define EOF (-1)
#endif

#ifndef NULL
#define NULL 0	/* Straight from Sun's stdio...  ugh. */
#endif

#ifdef AIX
#define VOLATILE volatile
#else
#define VOLATILE
#endif

typedef VOLATILE struct _XR_File {
    XR_Fildes		 _fd;
    int			 _cnt;
    char		*_base;
    char		*_ptr;
    short		 _flags;
    struct XR_MLRep	 _lock;
} XR_FILE;

#undef FILE
#define FILE XR_FILE

extern struct _XR_File _XR_iob[3];	/* The standard 3! */

#undef stdin
#undef stdout
#undef stderr

#define stdin  (&_XR_iob[0])
#define stdout (&_XR_iob[1])
#define stderr (&_XR_iob[2])

#define _SNFL	0x0001	/* Need flush */
#define _SLBF	0x0002	/* Line buffered if set, otherwise fully buffered */
#define _SERR	0x8000	/* Error */
#define _SDNF	0x4000	/* Don't Xfree() the buffer or the fp. */

extern int XR_fprintf(
#if NeedFunctionPrototypes
    XR_FILE *	/* file to write to */,
    char *	/* format string */,
    ...		/* Up to 8 arguments, assuming anything fits into an unsigned
		    long */
#endif
);

extern XR_FILE *XR_fopen(
#if NeedFunctionPrototypes
    char *	/* file to open */,
    char *	/* mode to open */
#endif
);

extern int XR_fclose(
#if NeedFunctionPrototypes
    XR_FILE *	/* stream to close */
#endif
);

extern int XR_fflush(
#if NeedFunctionPrototypes
    XR_FILE *	/* stream to flush */
#endif
);

extern void XR_rewind(
#if NeedFunctionPrototypes
    XR_FILE *	/* stream to rewind */
#endif
);

extern int XR_fread(
#if NeedFunctionPrototypes
    char *	/* buffer to read into */,
    size_t	/* size of each item */,
    size_t	/* number of items */,
    XR_FILE *	/* stream */
#endif
);

extern int XR_fwrite(
#if NeedFunctionPrototypes
    char *	/* buffer to write from */,
    size_t	/* size of each item */,
    size_t	/* number of items */,
    XR_FILE *	/* stream */
#endif
);

extern int XR_fputs(
#if NeedFunctionPrototypes
    char *	/* string */,
    XR_FILE *   /* stream */
#endif
);

char * XR_fgets(
#if NeedFunctionPrototypes
    char *	/* buffer for string */,
    size_t	/* length of buffer */,
    XR_FILE *	/* stream to read from */
#endif
);

int XR_fputc(
#if NeedFunctionPrototypes
    int		/* character to put */,
    XR_FILE *	/* stream to land it on */
#endif
);

int XR_fgetc(
#if NeedFunctionPrototypes
    XR_FILE *	/* stream to get character from */
#endif
);

#define fprintf	XR_fprintf
#define fopen	XR_fopen
#define fclose	XR_fclose
#define fflush	XR_fflush
#define fread	XR_fread
#define fwrite	XR_fwrite
#define fputs	XR_fputs
#define fgets	XR_fgets
#define fputc	XR_fputc
#undef	putc
#define putc	XR_fputc
#undef	getc
#define fgetc	XR_fgetc
#define getc	XR_fgetc

#endif	/* _PCR_H */	/* Do not add anything after this line */
