/*
 * Threas.h - translate OS dependent thread stuff into X thread stuff
 */

/* Copyright (C) 1992 by Digital Equipment Corporation and Xerox Corporation */

/*	All Rights Reserved
Permission to use, copy, modify, and distribute this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Digital and Xerox not
be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  Digital and
Xerox make no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.
*/

#ifndef THREADS_H
#define THREADS_H

#ifdef PTHREADS

/*
 * POSIX standard threads
 */

#include	<pthread.h>

/* variable types */
typedef pthread_mutex_t		Mutex;
typedef pthread_cond_t		Condition;
typedef struct
{
    pthread_once_t		block;
    int				flag;
}	OnceFlag;

/* Mutex functions */
#define InitMutex(mtx)		pthread_mutex_init((mtx),pthread_mutexattr_default)
#define LockMutex(mtx)		pthread_mutex_lock(mtx)
#define UnlockMutex(mtx)	pthread_mutex_unlock(mtx)
#define DestroyMutex(mtx)	pthread_mutex_destroy(mtx)

/* Condition Variable functions */
#define InitCondition(cv)	pthread_cond_init((cv),pthread_condattr_default)
#define WaitCondition(cv,mtx,time) \
if((time) != WAITFOREVER) \
{ struct timespec spec; \
  spec.tv_sec = (time)/1000; spec.tv_nsec = ((time)%1000)*1000000; \
  pthread_cond_timedwait((cv),(mtx),&spec); \
} else pthread_cond_wait((cv),(mtx));

#define NotifyCondition(cv)	pthread_cond_signal(cv)
#define BroadcastCondition(cv)	pthread_cond_broadcast(cv)
#define DestroyCondition(cv)	pthread_cond_destroy(cv)

/* Once functions */
#define DoOnce(once,func) \
if( ! (once)->flag ) { \
  pthread_once( &(once)->block, (func) ); \
  (once)->flag = True; }

/* static initialization */
#define NullMutex		{ 0 }
#define NullCondition		{ 0 }
#define OnceNotDone		{ pthread_once_init, False }

#endif	/* PTHREADS */

#ifdef PCR

/*
 * Xerox's in-house threads system
 */

/* variable types */
typedef struct XR_MLRep		Mutex;
typedef struct XR_CVRep		Condition;
typedef struct
{
    Mutex			lock;
    int				flag;
}	OnceFlag;

/* Mutex functions */

#define	InitMutex(mtx)		XR_InitializeMonitor((mtx))
#define DestroyMutex(mtx)
#define LockMutex(mtx)		XR_MonitorEntry((mtx))
#define UnlockMutex(mtx)	XR_MonitorExit((mtx))

/* Condition Variable functions */
#define	InitCondition(cond)	XR_InitializeCondition((cond), XR_WAIT_FOREVER)
#define DestroyCondition(cond)
#define WaitCondition(cond, mtx, timeout)	/* timeout in milliseconds */ \
	if ((timeout) != NOWAIT) {\
	    if ((timeout) != WAITFOREVER)\
		XR_SetTimeout((cond), ((timeout)+XR_MSECS_PER_TICK-1)/XR_MSECS_PER_TICK);\
	    XR_WaitCV((cond), (mtx));\
	    if (XR_GetTimeout((cond)) != XR_WAIT_FOREVER)\
		XR_DisableTimeout((cond));\
	};

#define	NotifyCondition(cond)		XR_Notify((cond))
#define	BroadcastCondition(cond)	XR_Broadcast((cond))

/* Once functions */
#define DoOnce(once,func) \
if( ! (once)->flag ) { \
  LockMutex( &(once)->lock ); \
  if( ! (once)->flag ) { (func)(); (once)->flag = True; } \
  UnlockMutex( &(once)->lock ); }

/* static initialization */
#define	NullMutex		XR_NULL_ML_REP
#define	NullCondition		XR_NULL_Condition_REP
#define OnceNotDone		{ NullMutex, False }

#endif	/* PCR */

/* Display locking (*dpy) */
#define LockDisplay(dpy)	LockMutex(&(dpy)->locks->general)
#define UnlockDisplay(dpy)	UnlockMutex(&(dpy)->locks->general)

#endif	/* THREADS_H    End Of File */
