/* File: PCR.c - last edit:
 * vsl      Thursday, February 13, 1992
 *
 * Created: 31-Oct-91 by Steve Watt
 *
 * Revision History:
 * 13-Feb-92 vsl	delete memchr prototype, and cast return value to (char *)    
 * 5-Dec-91  SEW	Added XR_XReadV and XR_XWriteV
 * 4-Dec-91  SEW	Added rewind
 * 4-Dec-91  CWS	added gethostbyname and changed name to PCR.c
 * 31-Oct-91 SEW	Created
 */
/* Copyright (C) 1992 by Digital Equipment Corporation and Xerox Corporation */

/*	All Rights Reserved
Permission to use, copy, modify, and distribute this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Digital and Xerox not
be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  Digital and
Xerox make no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.
*/


/*
 * include files
 */

#include "Xlib.h"
#include <varargs.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/uio.h>


#define SC_BUFSIZ  512
#define TRUE    1
#define FALSE   0

extern double strtod();

/*
   A truly simple thread-safe stdio.  There is NO support for seeking, or for
    switching between read and write on a particular file pointer.

   Well, hardly any support for seeking.  Rewind is now in.
*/


static int __sflags();
static int  _XR_refill();
static void _XR_fflush();

/* These buffers are appropriately locked when locking fp's. */
static char _XR_stdin_buf[BUFSIZ], _XR_stdout_buf[BUFSIZ],
	    _XR_stderr_buf[BUFSIZ];

XR_FILE _XR_iob[3] = {	/* Set up the big three */
    { XR_MSG_STDIN,  0, _XR_stdin_buf,  _XR_stdin_buf,  _SDNF|_SLBF, 
	XR_NULL_ML_REP },
    { XR_MSG_STDOUT, 0, _XR_stdout_buf, _XR_stdout_buf, _SDNF|_SLBF,
	 XR_NULL_ML_REP },
    { XR_MSG_STDOUT, 0, _XR_stderr_buf, _XR_stderr_buf, _SDNF|_SLBF,
	 XR_NULL_ML_REP }
};

/* This particular kludge made possible by the fact that PCR does not support
    more than 8 arguments to XR_FPrintF. */

int
XR_fprintf(fp, fmt, a0, a1, a2, a3, a4, a5, a6, a7)
    XR_FILE *fp;
    char *fmt;
    unsigned long a0, a1, a2, a3, a4, a5, a6, a7;
{
    char buf[BUFSIZ];
    int ret;

    XR_SPrintF(buf, fmt, a0, a1, a2, a3, a4, a5, a6, a7);

    ret = XR_fputs(buf, fp);

    return ret == EOF ? ret : strlen(buf);
}

XR_FILE *
XR_fopen(name, mode)
    char *name, *mode;
{
    XR_FILE *fp;
    int md_flag;

    if (!__sflags(mode, &md_flag))	/* build up a flag mask */
	return NULL;

    fp = (XR_FILE *)Xmalloc(sizeof(XR_FILE));	/* Get space for fp */
    if (fp == NULL)
	return NULL;

    fp->_base = (char *)Xmalloc(BUFSIZ);	/* and the buffer */
    if (fp->_base == NULL)
	return NULL;

    fp->_ptr = fp->_base;

    fp->_fd = XR_Open(name, md_flag, 0666);	/* Open it! */

    if (fp->_fd == XR_nullFildes)
	return NULL;

    XR_InitializeMonitor(&fp->_lock);	/* Set up the lock */

    fp->_cnt = 0;
    fp->_flags = 0;
    return fp;				/* and give it all back */
}

int
XR_fclose(fp)
    XR_FILE *fp;
{
    int ret;	/* Yeah, I know... who checks the return from fclose? */

    XR_MonitorEntry(&fp->_lock);

    if (fp->_flags & _SNFL)		/* flush it if needed */
	_XR_fflush(fp);

    ret = 0;

    if (fp->_flags & _SERR)		/* Any errors? */
	ret = EOF;

    if (XR_Close(fp->_fd) < 0)		/* close it */
	ret |= EOF;

    if (!(fp->_flags & _SDNF)) {	/* Are we allowed to free this? */
	Xfree(fp->_base);		/* let go of the buffer */
	Xfree(fp);
    }

    /* I'm leaving the lock set, in case anybody tries to use this FP
	before the memory gets re-used.  The thread will hang if it tries.  */

    return ret;
}

int
XR_fflush(fp)
    XR_FILE *fp;
{
    int ret;

    XR_MonitorEntry(&fp->_lock);	/* Lock it */

    if (fp->_flags & _SNFL)		/* flush if needed */
	_XR_fflush(fp);

    ret = fp->_flags & _SERR ? EOF : 0;	/* does anybody check the return
					    status of flush? */
    XR_MonitorExit(&fp->_lock);

    return ret;				/* probably not, but here it is. */
}

void
XR_rewind(fp)
    XR_FILE *fp;
{
    XR_MonitorEntry(&fp->_lock);	/* Lock it */

    if (fp->_flags & _SNFL)		/* flush if needed */
	_XR_fflush(fp);

    XR_LSeek( fp->_fd, 0, 0 );		/* back to the start! */

    fp->_cnt = 0;
    fp->_ptr = fp->_base;
    fp->_flags &= ~_SNFL;	/* Buffer contents don't need flush. */

    XR_MonitorExit(&fp->_lock);
}

int
XR_fread(buf, size, count, fp)
    char *buf;
    size_t size, count;
    XR_FILE *fp;
{
    size_t resid;
    char *p;
    int r;
    size_t total;

    if ((resid = count * size) == 0)
	return count;

    total = resid;
    p = buf;

    XR_MonitorEntry(&fp->_lock);		/* Lock it */

    while (resid > (r = fp->_cnt)) {	/* While there's more in the buffer */
	memcpy(p, fp->_ptr, r);		/* move it into the user buffer, */
	fp->_ptr += r;
	p += r;				/* update pointers and refill */
	resid -= r;
	/* refill does fp->_cnt = 0 */
	if (_XR_refill(fp)) {		/* read failed. */
	    XR_MonitorExit(&fp->_lock);
	    return (total - resid) / size;
	}
    }

    if (resid) {
	memcpy(p, fp->_ptr, resid);	/* copy remaining... */
	fp->_cnt -= resid;
	fp->_ptr += resid;
    }

    XR_MonitorExit(&fp->_lock);

    return count;
}

int
XR_fwrite(buf, size, count, fp)
    char *buf;
    size_t size, count;
    XR_FILE *fp;
{
    size_t resid;
    char *p;
    int w;
    size_t total;

    if ((resid = count * size) == 0)
	return count;

    total = resid;
    p = buf;

    XR_MonitorEntry(&fp->_lock);

    while (resid > (w = (BUFSIZ - fp->_cnt))) {
	memcpy(fp->_ptr, p, w);		/* fill the buffer */
	fp->_flags |= _SNFL;
	resid -= w;
	p += w;
	/* flush does fp->_cnt = 0 and fp->_ptr = fp->_base */
	_XR_fflush(fp);
	if (fp->_flags & _SERR) {	/* write failed */
	    XR_MonitorExit(&fp->_lock);
	    return (total - resid) / size;
	}
    }

    if (resid) {
	memcpy(fp->_ptr, p, resid);	/* copy remaining. */
	fp->_flags |= _SNFL;
	fp->_ptr += resid;
	fp->_cnt += resid;
	if ((fp->_flags & _SLBF) && (char *)memchr(p, '\n', resid))
	    _XR_fflush(fp);
	    if (fp->_flags & _SERR)		/* write failed */
		return (total - resid) / size;
    }

    XR_MonitorExit(&fp->_lock);

    return count;
}

int
XR_fputs(string, fp)
    char *string;
    XR_FILE *fp;
{
    int len = strlen(string);
    int ret;

    XR_MonitorEntry(&fp->_lock);	/* lock it */

    if (len + fp->_cnt >= BUFSIZ) {
	if (fp->_flags & _SNFL)
	    _XR_fflush(fp);		/* Empty the buffer, if needed */
    }

    if (len >= BUFSIZ) {		/* If it's going to (over?)fill the
					    (now empty) buffer, just write it
					    out. */
	ret = XR_Write(fp->_fd, string, len);
	/* if XR_Write returns a partial write, tough.  This may need to
	    be fixed.*/

	if (ret != len)
	    fp->_flags |= _SERR;

	len = 0;			/* Let the rest of the code know it's
					    gone. */
    }

    /* By the time we're here, if len is not 0, we know it'll fit into the
	buffer. */

    if (len) {
	memcpy(fp->_ptr, string, len);
	fp->_ptr += len;		/* Move it, update pointers, set */
	fp->_cnt += len;		/*  flags */
	fp->_flags |= _SNFL;

	/* Poor man's line buffering:  If there's a newline anywhere in the
	    string, flush it */
	if ((fp->_flags & _SLBF) && strchr(string, '\n'))
	    _XR_fflush(fp);
    }

    ret = fp->_flags & _SERR ? EOF : 0;

    XR_MonitorExit(&fp->_lock);

    return ret;
}

char *
XR_fgets(buf, n, fp)
    char *buf;
    size_t n;
    XR_FILE *fp;
{
    char *s, *t, *p;
    int len;

    if (n < 2)    /* Sanity check. */
	return NULL;

    s = buf;
    n--;	/* leave space for NULL */

    XR_MonitorEntry(&fp->_lock);	/* lock it */

    do {				 /* If the buffer is empty, fill it */
	if ((len = fp->_cnt) == 0) {
	    if (_XR_refill(fp)) {	 /* EOF: stop with partial or no line */
		if (s == buf)
		    return NULL;

		break ;
	    }

	    len = fp->_cnt;
	}

	p = fp->_ptr;
	/* Look through n bytes of buffer for '\n'.  If found, copy up to it
	    and leave.  Otherwise, copy whole chunk and leave. */

	if (len > n)
	    len = n;

	t = (char *)memchr(p, '\n', len);
	if (t) {
	    len  = ++t - p;		/* found '\n' */
	    fp->_cnt -= len;		/* fix up pointers */
	    fp->_ptr = t;
	    memcpy(s, p, len);		/* copy it out */
	    XR_MonitorExit(&fp->_lock);
	    s[len] = 0;
	    return buf;
	}

	fp->_cnt -= len;
	fp->_ptr += len;
	memcpy(s, p, len);		/* copy this hunk */
	s += len;
    } while (n -= len);

    XR_MonitorExit(&fp->_lock);

    *s = 0;

    return buf;
}

int
XR_fputc(c, fp)
    int c;
    XR_FILE *fp;
{
    int ret;
    XR_MonitorEntry(&fp->_lock);	/* Lock it */

    if (fp->_cnt >= BUFSIZ) {
	if (fp->_flags & _SNFL)
	    _XR_fflush(fp);		/* Empty the buffer, if needed */
    }

    fp->_cnt++;
    *(fp->_ptr++) = c;
    fp->_flags |= _SNFL;		/* Shove the char into the buffer,
					    make sure the dirty flag is set */

    if ((fp->_flags & _SLBF) && (c == '\n'))
	_XR_fflush(fp);

    ret = fp->_flags & _SERR ? EOF : c;

    XR_MonitorExit(&fp->_lock);

    return ret;
}

int
XR_fgetc(fp)
    XR_FILE *fp;
{
    int ret;

    XR_MonitorEntry(&fp->_lock);	/* Lock it */

    ret = 0;
    if (fp->_cnt == 0) {		/* Need a refill? */
	if (_XR_refill(fp))		/* self-serv... ;) */
	    ret = EOF;
    }

    if (!ret) {
	ret = *fp->_ptr++;		/* grab a char */
	fp->_cnt--;
    }

    XR_MonitorExit(&fp->_lock);

    return ret;
}

/* Assumes that fp is already locked, and that the buffer does in fact need
    refilling. */

static int
_XR_refill(fp)
    XR_FILE *fp;
{
    int n;

    n = XR_Read(fp->_fd, fp->_base, BUFSIZ);
    fp->_ptr = fp->_base;
    fp->_flags &= ~_SNFL;	/* Buffer contents don't need flush. */
    fp->_cnt = 0;

    if (n <= 0)
	return EOF;

    fp->_cnt = n;
    return 0;
}

/* Assumes that fp is already locked, and that the buffer does in fact need
    to be emptied. */

static void
_XR_fflush(fp)
    XR_FILE *fp;
{
    int n;

    n = XR_Write(fp->_fd, fp->_base, fp->_cnt);

    if (n != fp->_cnt)
	fp->_flags |= _SERR;

    fp->_cnt = 0;
    fp->_ptr = fp->_base;
    fp->_flags &= ~_SNFL;	/* Buffer contents don't need flush. */
}

/*
 * Return the (stdio) flags for a given mode.  Store the flags
 * to be passed to an open() syscall through *optr.
 * Return 0 on error.
 */
static int
__sflags(mode, optr)
    register char *mode;
    int *optr;
{
    register int m, o;

    switch (*mode++) {

	case 'r':	/* open for reading */
	    m = O_RDONLY;
	    o = 0;
	    break;

	case 'w':	/* open for writing */
	    m = O_WRONLY;
	    o = O_CREAT | O_TRUNC;
	    break;

	case 'a':	/* open for appending */
	    m = O_WRONLY;
	    o = O_CREAT | O_APPEND;
	    break;

	default:	/* illegal mode */
    	    XR_errno = EINVAL;
	    return 0;
    }

    /* [rwa]\+ or [rwa]b\+ means read and write */
    if (*mode == '+' || (*mode == 'b' && mode[1] == '+'))
	m = O_RDWR;

    *optr = m | o;
    return 1;
}


/*
 * XR_gethostbyname
 * Created by Carl Schmidtmann  10-Jan-92
 *
 * This routine was copied from Pilot and modified to emulate the real
 * gethostbyname fucntion.
 *
 * Revision History:
 *
 * 10-Jan-92	CWS	Original version
 */

/*
 * start of XR_gethostbyname routine
 */

#define SPAWN_FAILED	256
#define MAXNAMELEN	256L
#define GETHOSTENV	"GETHOSTCMD"

static char	getHostCmd[] = "/usr/bin/xerox/gethostbyname";
static char	addr_str[MAXNAMELEN];
static char	*addr_array[] = { addr_str };

static char	h_name[MAXNAMELEN];
static struct hostent	hostdata = { h_name, NULL, 0, 0, addr_array };

extern char *XR_getenv();

static char *
line_copy( buf1, buf2, len )	/* copy one line of buf2 to buf1 */
    char	*buf1;
    char	*buf2;
    int		len;
{
    if(( buf1 == NULL ) || ( buf2 == NULL ))
    {
	return( NULL );
    }
    while( *buf2 && ( *buf2 != '\n' ) && --len )
    {
	*buf1++ = *buf2++;
    }
    if(( *buf2 && len ) || ( *buf2 == '\n' ))
    {
	buf2++;
    }
    *buf1 = NULL;
    return( buf2 );

}	/* end of line_copy function */

struct hostent *
xr_gethostbyname( hostname )
    char *hostname;
{
    char	*command;
    char	cmdline[MAXNAMELEN];
    char	line[MAXNAMELEN];
    char	buffer[2*MAXNAMELEN];
    char	*pos;
    int		length;
    XR_Fildes	fd;
    unsigned long	addr;
  
/*
 * create the command line
 */

    if(( command = XR_getenv( GETHOSTENV )) == NULL )
    {
	command = getHostCmd;
    }
    strcpy( cmdline, command );
    strcat( cmdline, " " );
    strcat( cmdline, hostname );

/*
 * get the host entry and fill in the data structure
 */

    fd = XR_PPOpen( cmdline );
    if( fd == -1 )
    {
	return( (struct hostent*)NULL );
    }

    length = XR_Read( fd, buffer, (2*MAXNAMELEN)-1 );
    if( length <= 0 )
    {
	return( (struct hostent*)NULL );
    }
    buffer[length] = NULL;	/* just to make sure */

    pos = line_copy( hostdata.h_name, buffer, MAXNAMELEN );
    pos = line_copy( line, pos, MAXNAMELEN );
    hostdata.h_addrtype = atoi( line );
    pos = line_copy( line, pos, MAXNAMELEN );
    hostdata.h_length = atoi( line );
    pos = line_copy( line, pos, MAXNAMELEN );
    addr = strtol( line, (char **)NULL, 0 );
    bcopy( (char *)(&addr), hostdata.h_addr, hostdata.h_length );

    if( strlen( line ) == 0 )
    {
	return( (struct hostent*)NULL );
    }
    else
    {
	return( &hostdata );
    }

}	/* end of XR_gethostbyname function */


static int
convert(fmt)
    char	*fmt;
{
    int		val;
    char	fctl;
 
    val=0;
    while (isdigit(*fmt))
	val = (10 * val) + *fmt++ -'0';

    if (!val) {
	if (*fmt == 'l' || *fmt == 'L')
	    fmt++;

	switch (fctl = *fmt) {
	    case 'd':
	    case 'D':
		val = 11;
		break;
 
	    case 'i':
	    case 'I':
		val = 13;
                break;
 
	    case 'u':
	    case 'U':
		val = 10;
                break;
 
	    case 'o':
	    case 'O':
		val = 13;
                break;
 
	    case 'x':
	    case 'X':
                val = 8;
                break;
 
	    case 'c':
	    case 'C':
                val = 1;
                break;
 
	    default:
                val = SC_BUFSIZ - 2;
                break;
	}
    }

    return ((val > SC_BUFSIZ-2) ? SC_BUFSIZ-2 : val);
}

int
XR_sscanf(va_alist)
va_dcl			/* NO SEMICOLON!  That's how the macro works. :( */
{
    va_list	args;
    char	*buf;
    char	*fmt;

    char	temp[SC_BUFSIZ];
    char	*str, *ptr, *tptr;
    int		maxchars, nargs, i;
    int		ignore, space, lng;
    char	fctl;

    va_start(args);

    buf = va_arg(args, char *);
    fmt = va_arg(args, char *);

    nargs = 0;
    ignore = FALSE;
    space = FALSE;
    lng = FALSE;

    str = buf;

    while (*fmt) {
	if (*fmt == '%')  {
	    ++fmt;

	   if  (*fmt == '*') {
		ignore = TRUE;
		++fmt;
            }

	    maxchars = convert(fmt);

	    while (isdigit(*fmt)) /* skip numeric, convert() handled it */
		fmt++;

	    ptr = temp;
	    tptr = str;
	    while (*ptr = *tptr)
		if (isspace(*ptr))
		    break ;
		else
		    tptr++, ptr++;

	    *ptr = '\0';

	    if (*fmt == 'l' || *fmt == 'L') {
		fmt++;
		lng++;
	    }

	    switch (fctl = *fmt++) {
		case 'd':
		    if (lng)
			*(va_arg(args, long *)) = strtol(temp, &ptr, 10);
		    else
			*(va_arg(args, int *)) = (int)strtol(temp, &ptr, 10);
		    break ;

		case 'i':
		    if (lng)
			*(va_arg(args, long *)) = strtol(temp, &ptr, 10);
		    else
			*(va_arg(args, int *)) = (int)strtol(temp, &ptr, 10);
		    break ;

		case 'u':
		    if (*temp =='-')
			*temp=' ';

		    if (lng)
			*(va_arg(args, unsigned long *)) =
			    strtol(temp, &ptr, 10);
		    else
			*(va_arg(args, unsigned *)) =
			    (int)strtol(temp, &ptr, 10);
		    break ;

		case 'o':
		    if (lng)
			*(va_arg(args, long *)) = strtol(temp, &ptr, 8);
		    else
			*(va_arg(args, int *)) = (int)strtol(temp, &ptr, 8);
		    break ;

		case 'x':
		    if (lng)
			*(va_arg(args, long *)) = strtol(temp, &ptr, 16);
		    else
			*(va_arg(args, int *)) = (int)strtol(temp, &ptr, 16);
		    break ;


		case 'f':
		    if (lng)
			*(va_arg(args, double *)) = strtod(temp, &ptr);
		    else
			*(va_arg(args, float *)) = (float)strtod(temp, &ptr);
		    break ;

		case 'c':
		    *(va_arg(args, char *)) = *temp;
		    ptr = temp + 1;
		    break;

		case 's':
		    strcpy (va_arg(args, char *), temp);
		    break;

		default:
		    *(va_arg(args, char *)) = '\0';
		    break;
	    }  /*end switch */

	    ++nargs;
	    fmt--;
	    str += ptr - temp;	/* bump source string pointer */
	}   /* end if '%' */
	else {
	    if (isspace(*fmt)) {
		while (isspace(*fmt)) {
		    while (isspace(*str))
			str++;
		    fmt++;
		}
		fmt--;
	    }
	    else {
	    /* If it's not a format spec, it has to match */
		if (*str != *fmt)
		    return nargs;
		else
		    str++;
	    }
	}
	fmt++;
    } /* end while */

    return nargs;
}   /* end scanf */


#if !defined(lint) && defined(SCCSIDS)
static	char sccsid[] = "@(#)qsort.c 1.1 90/10/29 SMI"; /* from UCB 4.2 83/03/089 */
#endif

/*
 * qsort.c:
 * Our own version of the system qsort routine which is faster by an average
 * of 25%, with lows and highs of 10% and 50%.
 * The THRESHold below is the insertion sort threshold, and has been adjusted
 * for records of size 48 bytes.
 * The MTHREShold is where we stop finding a better median.
 */

#define		THRESH		4		/* threshold for insertion */
#define		MTHRESH		6		/* threshold for median */

struct qs {
    int		(*qcmp)();		/* the comparison routine */
    int		qsz;			/* size of each record */
    int		thresh;			/* THRESHold in chars */
    int		mthresh;		/* MTHRESHold in chars */
};

/*
 * qsort:
 * First, set up some global parameters for qst to share.  Then, quicksort
 * with qst(), and then a cleanup insertion sort ourselves.  Sound simple?
 * It's not...
 */

void XR_qsort(base, n, size, compar)
	char	*base;
	int	n;
	int	size;
	int	(*compar)();
{
	struct qs qs;
	register char c, *i, *j, *lo, *hi;
	char *min, *max;

	if (n <= 1)
		return;
	qs.qsz = size;
	qs.qcmp = compar;
	qs.thresh = qs.qsz * THRESH;
	qs.mthresh = qs.qsz * MTHRESH;
	max = base + n * qs.qsz;
	if (n >= THRESH) {
		qst(base, max, &qs);
		hi = base + qs.thresh;
	} else {
		hi = max;
	}
	/*
	 * First put smallest element, which must be in the first THRESH, in
	 * the first position as a sentinel.  This is done just by searching
	 * the first THRESH elements (or the first n if n < THRESH), finding
	 * the min, and swapping it into the first position.
	 */
	for (j = lo = base; (lo += qs.qsz) < hi; )
		if (qs.qcmp(j, lo) > 0)
			j = lo;
	if (j != base) {
		/* swap j into place */
		for (i = base, hi = base + qs.qsz; i < hi; ) {
			c = *j;
			*j++ = *i;
			*i++ = c;
		}
	}
	/*
	 * With our sentinel in place, we now run the following hyper-fast
	 * insertion sort.  For each remaining element, min, from [1] to [n-1],
	 * set hi to the index of the element AFTER which this one goes.
	 * Then, do the standard insertion sort shift on a character at a time
	 * basis for each element in the frob.
	 */
	for (min = base; (hi = min += qs.qsz) < max; ) {
		while (qs.qcmp(hi -= qs.qsz, min) > 0)
			/* void */;
		if ((hi += qs.qsz) != min) {
			for (lo = min + qs.qsz; --lo >= min; ) {
				c = *lo;
				for (i = j = lo; (j -= qs.qsz) >= hi; i = j)
					*i = *j;
				*i = c;
			}
		}
	}
}

/*
 * qst:
 * Do a quicksort
 * First, find the median element, and put that one in the first place as the
 * discriminator.  (This "median" is just the median of the first, last and
 * middle elements).  (Using this median instead of the first element is a big
 * win).  Then, the usual partitioning/swapping, followed by moving the
 * discriminator into the right place.  Then, figure out the sizes of the two
 * partions, do the smaller one recursively and the larger one via a repeat of
 * this code.  Stopping when there are less than THRESH elements in a partition
 * and cleaning up with an insertion sort (in our caller) is a huge win.
 * All data swaps are done in-line, which is space-losing but time-saving.
 * (And there are only three places where this is done).
 */

static
qst(base, max, qsp)
	char *base, *max;
	struct qs *qsp;
{
	register char c, *i, *j, *jj;
	register int ii;
	char *mid, *tmp;
	int lo, hi;

	/*
	 * At the top here, lo is the number of characters of elements in the
	 * current partition.  (Which should be max - base).
	 * Find the median of the first, last, and middle element and make
	 * that the middle element.  Set j to largest of first and middle.
	 * If max is larger than that guy, then it's that guy, else compare
	 * max with loser of first and take larger.  Things are set up to
	 * prefer the middle, then the first in case of ties.
	 */
	lo = max - base;		/* number of elements as chars */
	do	{
		mid = i = base + qsp->qsz * ((lo / qsp->qsz) >> 1);
		if (lo >= qsp->mthresh) {
			j = (qsp->qcmp((jj = base), i) > 0 ? jj : i);
			if (qsp->qcmp(j, (tmp = max - qsp->qsz)) > 0) {
				/* switch to first loser */
				j = (j == jj ? i : jj);
				if (qsp->qcmp(j, tmp) < 0)
					j = tmp;
			}
			if (j != i) {
				ii = qsp->qsz;
				do	{
					c = *i;
					*i++ = *j;
					*j++ = c;
				} while (--ii);
			}
		}
		/*
		 * Semi-standard quicksort partitioning/swapping
		 */
		for (i = base, j = max - qsp->qsz; ; ) {
			while (i < mid && qsp->qcmp(i, mid) <= 0)
				i += qsp->qsz;
			while (j > mid) {
				if (qsp->qcmp(mid, j) <= 0) {
					j -= qsp->qsz;
					continue;
				}
				tmp = i + qsp->qsz;	/* value of i after swap */
				if (i == mid) {
					/* j <-> mid, new mid is j */
					mid = jj = j;
				} else {
					/* i <-> j */
					jj = j;
					j -= qsp->qsz;
				}
				goto swap;
			}
			if (i == mid) {
				break;
			} else {
				/* i <-> mid, new mid is i */
				jj = mid;
				tmp = mid = i;	/* value of i after swap */
				j -= qsp->qsz;
			}
		swap:
			ii = qsp->qsz;
			do	{
				c = *i;
				*i++ = *jj;
				*jj++ = c;
			} while (--ii);
			i = tmp;
		}
		/*
		 * Look at sizes of the two partitions, do the smaller
		 * one first by recursion, then do the larger one by
		 * making sure lo is its size, base and max are update
		 * correctly, and branching back.  But only repeat
		 * (recursively or by branching) if the partition is
		 * of at least size THRESH.
		 */
		i = (j = mid) + qsp->qsz;
		if ((lo = j - base) <= (hi = max - i)) {
			if (lo >= qsp->thresh)
				qst(base, j, qsp);
			base = i;
			lo = hi;
		} else {
			if (hi >= qsp->thresh)
				qst(i, max, qsp);
			max = j;
		}
	} while (lo >= qsp->thresh);
}

#ifdef OSF1

int XR_XReadV( fildes, iov, iovCnt )
    XR_Fildes		fildes;
    struct iovec	*iov;
    int			iovCnt;
{
    int nbytes = 0;
    int index = 0;
    int result;

    while ( index < iovCnt )
    {
	result = XR_Read( fildes, iov[index].iov_base, iov[index].iov_len );
	if ( result < 0 )
	    return result;	/**  Error occurred on read **/
	if ( result == 0 )
	    return result;	/**  EOF reached  **/
	nbytes += result;
	index++;
    }
    return nbytes;
}

XR_XWriteV( fildes, iov, iovCnt )
    XR_Fildes		fildes;
    struct iovec	*iov;
    int			iovCnt;
{
    int nbytes = 0;
    int index = 0;
    int result;

    if ( iovCnt < 1 || iovCnt > 16 )
	return -1;	/** errno should be set to EINVAL, but how? **/
    while ( index < iovCnt )
    {
	if ( iov[index].iov_len < 0 )
	    return -1;	/** errno should be set to EINVAL, but how? **/
	result = XR_Write( fildes, iov[index].iov_base, iov[index].iov_len );
	if ( result < 0 )
	    return result;	/**  Error occurred on write **/
	nbytes += result;
	index++;
    }
    return nbytes;
}

#endif	/* OSF1 */

long _XEventToMask( index )
    int		index;
{
    extern long _Xevent_to_mask[];
    return( _Xevent_to_mask[index] );
}

