/* $XConsortium: XErrHndlr.c,v 11.15 91/01/24 11:10:03 rws Exp $ */
/* Copyright    Massachusetts Institute of Technology    1986	*/

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

#include "Xlibint.h"

#ifdef MULTI_THREAD

/*
 * start of _XAddHandler routine
 */

XHandlerStruct *
_XAddHandler( dpy, list, handler, data )
    Display		*dpy;
    int			list;
    XErrorHandler	handler;
    XPointer		data;
{
    XHandlerList	handler_str;

/*
 * initialize the handler structure
 */

    handler_str = (XHandlerList)Xmalloc( sizeof( XHandlerStruct ));
    handler_str->next = NULL;
    handler_str->func = handler;
    handler_str->data = data;

/*
 * add the structure to the appropriate list
 */

    LockMutex( &dpy->locks->error_list );
    switch( list )
    {
      case INTERNAL_LIST:
	handler_str->next = dpy->internal_handlers;
	dpy->internal_handlers = handler_str;
	break;

      case EXTERNAL_LIST:
	handler_str->next = dpy->external_handlers;
	dpy->external_handlers = handler_str;
	break;

      case FATAL_LIST:
	handler_str->next = dpy->fatal_handlers;
	dpy->fatal_handlers = handler_str;
	break;

    }	/* end of switch statement */
    UnlockMutex( &dpy->locks->error_list );

    return( handler_str );

}	/* end of _XAddHandler routine */

/*
 * start of _XRemoveHandler routine
 */

Bool
_XRemoveHandler( dpy, list, handler )
    Display		*dpy;		/* Display structure */
    int			list;		/* which error handler list? */
    XHandlerList	handler;	/* which handler in the list? */
{
    XHandlerList	*handler_list;
    XHandlerList	curr_handler;
    XHandlerList	prev_handler;

    if( handler == NULL )	/* is there anything to do? */
    {
	return( FALSE );
    }

    LockMutex( &dpy->locks->error_list );

/*
 * get the proper list to search
 */

    switch( list )
    {
      case INTERNAL_LIST:
	handler_list = &dpy->internal_handlers;
	break;
      case EXTERNAL_LIST:
	handler_list = &dpy->external_handlers;
	break;
      case FATAL_LIST:
	handler_list = &dpy->fatal_handlers;
	break;
      default:
	return( FALSE );
    }	/* end of switch statement */

/*
 * find and remove the handler structure
 */

    curr_handler = *handler_list;
    prev_handler = NULL;
    while( curr_handler != NULL )
    {
	if( curr_handler == handler )
	{
	    if( prev_handler == NULL )
	    {
		*handler_list = curr_handler->next;
	    }
	    else
	    {
		prev_handler->next = curr_handler->next;
	    }
	    break;
	}
	prev_handler = curr_handler;
	curr_handler = curr_handler->next;
    }	/* end of while loop */

    UnlockMutex( &dpy->locks->error_list );
    Xfree( handler );
    return( TRUE );

}	/* end of _XRemoveHandler routine */
    
/*
 * start of XAddErrorHandler routine
 */

XHandlerStruct *
XAddErrorHandler( dpy, handler, data )
    Display		*dpy;		/* Display structure */
    XErrorHandler	handler;	/* error handler routine */
    XPointer		data;		/* client data */
{
    return( _XAddHandler( dpy, EXTERNAL_LIST, handler, data ));

}	/* end of XAddErrorHandler routine */

/*
 * start of XRemoveErrorHandler routine
 */

Bool
XRemoveErrorHandler( dpy, handler )
    Display		*dpy;		/* Display structure */
    XHandlerList	handler;	/* error handler structure */
{
    return( _XRemoveHandler( dpy, EXTERNAL_LIST, handler ));

}	/* end of XRemoveErrorHandler routine */

/*
 * start of XAddIOErrorHandler routine
 */

XHandlerStruct *
XAddIOErrorHandler( dpy, handler, data )
    Display		*dpy;		/* Display structure */
    XErrorHandler	handler;	/* error handler routine */
    XPointer		data;		/* client data */
{
    return( _XAddHandler( dpy, FATAL_LIST, handler, data ));

}	/* end of XAddIOErrorHandler routine */

/*
 * start of XRemoveIOErrorHandler routine
 */

Bool
XRemoveIOErrorHandler( dpy, handler )
    Display		*dpy;		/* Display structure */
    XHandlerList	handler;	/* error handler structure */
{
    return( _XRemoveHandler( dpy, FATAL_LIST, handler ));

}	/* end of XRemoveIOErrorHandler routine */

Mutex _Xerror_function_mutex;

void
_XInitErrFuncMutex()
{
    InitMutex( &_Xerror_function_mutex );
}

#endif	/* MULTI_THREAD */

extern int _XDefaultError();
extern int _XDefaultIOError();
/* 
 * XErrorHandler - This procedure sets the X non-fatal error handler
 * (_XErrorFunction) to be the specified routine.  If NULL is passed in
 * the original error handler is restored.
 */
 
XErrorHandler XSetErrorHandler(handler)
    register XErrorHandler handler;
{
    int (*oldhandler)();

    LockMutex( &_Xerror_function_mutex );

    oldhandler = _XErrorFunction;

    if (!oldhandler)
	oldhandler = _XDefaultError;

    if (handler != NULL) {
	_XErrorFunction = handler;
    }
    else {
	_XErrorFunction = _XDefaultError;
    }

    UnlockMutex( &_Xerror_function_mutex );

    return oldhandler;
}

/* 
 * XIOErrorHandler - This procedure sets the X fatal I/O error handler
 * (_XIOErrorFunction) to be the specified routine.  If NULL is passed in 
 * the original error handler is restored.
 */
 
extern int _XIOError();
XIOErrorHandler XSetIOErrorHandler(handler)
    register XIOErrorHandler handler;
{
    int (*oldhandler)();

    LockMutex( &_Xerror_function_mutex );

    oldhandler = _XIOErrorFunction;

    if (!oldhandler)
	oldhandler = _XDefaultIOError;

    if (handler != NULL) {
	_XIOErrorFunction = handler;
    }
    else {
	_XIOErrorFunction = _XDefaultIOError;
    }

    UnlockMutex( &_Xerror_function_mutex );

    return oldhandler;
}

