/* File: XEventsMQ.c - last edit:
 * Michael X. Tao	Mon Nov 25 11:58:23 1991
 * MXT		Thu Nov 21 16:31:09 1991
 */

/* Copyright (C) 1992 by Digital Equipment Corporation and Xerox Corporation */

/*	All Rights Reserved
Permission to use, copy, modify, and distribute this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Digital and Xerox not
be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  Digital and
Xerox make no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.
*/

#ifdef	MULTI_THREAD
#define NEED_EVENTS
#define NEED_REPLIES
#include "Xlibint.h"
#include <sys/time.h>

#define AllPointers (PointerMotionMask|PointerMotionHintMask|ButtonMotionMask)
#define AllButtons (Button1MotionMask|Button2MotionMask|Button3MotionMask|\
		    Button4MotionMask|Button5MotionMask)

/* 
 * This routine creates an event queue and returns a handle to it.
 */
XQueue *XCreateEventQueue(dpy, valuemask, queue_properties)
	register Display *dpy;
	register unsigned long valuemask;
	register XQueueProp *queue_properties;
	
{
	register XQueue *queue;		/* event queue to be created */

	if ((queue  = (XQueue *)Xmalloc(sizeof(XQueue))) == NULL) {
	    UnlockDisplay(dpy);
	    return queue;			/* fail to create a queue */
	}
	queue->head = queue->tail = NULL;
	queue->qlen = 0;
	queue->properties.mask = (valuemask & QPMask) ?
				queue_properties->mask : 0xffffffff;
	queue->properties.predicate = (valuemask & QPPredicate) ?
				queue_properties->predicate : NULL;
	queue->properties.pred_data = (valuemask & QPPredicateData) ?
				queue_properties->pred_data : NULL;

	InitMutex(&queue->lock);
	InitCondition(&queue->condvar);
	queue->waiting = False;

	LockDisplay(dpy);		/* put queue at the 2nd spot */
	queue->next = dpy->event_que->next;
	dpy->event_que->next = queue;
	UnlockDisplay(dpy);
	return queue;
}	/* end of XCreateEventQueue routine */

/*
 * Free all the events in the event queue and set its qlen to zero.
 * Assume the caller is holding the queue lock.
 */
_XFreeEvents(event_queue)
	register XQueue *event_queue;
{
	register _XQEvent *qelt;

	for (qelt = (_XQEvent *)event_queue->head;
	    (_XQEvent *)event_queue->head;
	    qelt = (_XQEvent *)event_queue->head) {
		event_queue->head = (XPointer)((_XQEvent *)event_queue->head)->next;
		_XFreeBlock((_XQBlock *)qelt);
	};
	event_queue->tail = NULL;
	event_queue->qlen = 0;
};

/* 
 * Destory an event queue previously created by XCreateEventQueue.
 * Note that the default (first) event queue cannot be removed.
 */
Bool XDestroyEventQueue(dpy, event_queue)
	Display* dpy;
	register XQueue *event_queue;	/* queue to be deleted */
{
	register XQueue *prev_queue = dpy->event_que;

	if (dpy->event_que == event_queue)
	    return False;		/* can't delete the default queue */

	LockDisplay(dpy);
	while (prev_queue->next != event_queue) {
	    prev_queue = prev_queue->next;
	    if (prev_queue == NULL) {
		UnlockDisplay(dpy);
		return False;		/* fail to find the queue in the list */
	    }
	}

	LockMutex(&event_queue->lock);	/* can't receive anymore event */
	prev_queue->next = event_queue->next;
	UnlockDisplay(dpy);

	event_queue->next = NULL;
	_XFreeEvents(event_queue);
	event_queue->qlen = -1;		/* for client debugging */
	
	Xfree(event_queue);
	/* don't bother to unlock event_queue */

	return True;
}	/* end of XDestroyEventQueue routine */

/* 
 * This routine chages an event queue's properties
 * and returns whether the change is successful.
 */
Bool XChangeEventQueueProperties(dpy, event_queue, valuemask, queue_properties)
	register Display *dpy;
	XQueue *event_queue;		/* queue's properties to changed */
	register unsigned long valuemask;
	register XQueueProp *queue_properties;
	
{
	register XQueue *queue = dpy->event_que;

	LockDisplay(dpy);
	while (queue != event_queue) {
	    queue = queue->next;
	    if (queue == NULL) {
		UnlockDisplay(dpy);
		return False;		/* fail to find the queue in the list */
	    }
	}

	LockMutex(&queue->lock);
	UnlockDisplay(dpy);

	if (valuemask & QPMask)
	   queue->properties.mask = queue_properties->mask;
	if (valuemask & QPPredicate)
	   queue->properties.predicate = queue_properties->predicate;
	if (valuemask & QPPredicateData)
	   queue->properties.pred_data = queue_properties->pred_data;

	UnlockMutex(&queue->lock);
	return True;
};

/* 
 * This routine gets an event queue's properties
 * and returns whether the queue is found.
 */
Bool XGetEventQueueProperties(dpy, event_queue, valuemask, queue_properties)
	register Display *dpy;
	XQueue *event_queue;		/* queue's properties to changed */
	register unsigned long valuemask;
	register XQueueProp *queue_properties;
	
{
	register XQueue *queue = dpy->event_que;

	LockDisplay(dpy);
	while (queue != event_queue) {
	    queue = queue->next;
	    if (queue == NULL) {
		UnlockDisplay(dpy);
		return False;		/* fail to find the queue in the list */
	    }
	}

	LockMutex(&queue->lock);
	UnlockDisplay(dpy);

	if (valuemask & QPMask)
	   queue_properties->mask = queue->properties.mask;
	if (valuemask & QPPredicate)
	   queue_properties->predicate = queue->properties.predicate;
	if (valuemask & QPPredicateData)
	   queue_properties->pred_data = queue->properties.pred_data;

	UnlockMutex(&queue->lock);
	return True;
};

/* 
 * Check existing events in queue to find if any match.  If so, return.
 * If not, flush buffer and see if any more events are readable. If one
 * matches, return.  If all else fails, tell the user no events found.
 */
Bool XCheckIfEventMQ (dpy, event_queue, event, predicate, arg)
	Display* dpy;
        register XQueue *event_queue;
	XEvent *event;		/* XEvent to be filled in. */
	XPredicate predicate;	/* function to call */
	XPointer *arg;
{
	register _XQEvent *prev, *qelt;
	int n;			/* time through count */

	prev = NULL;
	for (n = 3; --n >= 0;) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((*predicate)(dpy, &qelt->event, arg)) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }
	    UnlockMutex(&event_queue->lock);
	    switch (n) {
	      case 2:
		_XEventsQueued(dpy, event_queue, QueuedAfterReading);
		break;
	      case 1:
		XFlush(dpy);
		break;
	    }
	}
	return False;
}

extern long _Xevent_to_mask[];

/* 
 * Check existing events in queue to find if any match.  If so, return.
 * If not, flush buffer and see if any more events are readable. If one
 * matches, return.  If all else fails, tell the user no events found.
 */
Bool XCheckMaskEventMQ (dpy, event_queue, mask, event)
	Display* dpy;
	register XQueue *event_queue;
	register long mask;	/* Selected event mask. */
 	XEvent *event;		/* XEvent to be filled in. */
{
	register _XQEvent *prev, *qelt;
	int n;			/* time through count */

	prev = NULL;
	for (n = 3; --n >= 0;) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		 qelt;
		 prev = qelt, qelt = qelt->next) {
		if ((qelt->event.type < LASTEvent) &&
		    (_Xevent_to_mask[qelt->event.type] & mask) &&
		    ((qelt->event.type != MotionNotify) ||
		     (mask & AllPointers) ||
		     (mask & AllButtons & qelt->event.xmotion.state))) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }
	    UnlockMutex(&event_queue->lock);
	    switch (n) {
	      case 2:
		_XEventsQueued(dpy, event_queue, QueuedAfterReading);
		break;
	      case 1:
		XFlush(dpy);
		break;
	    }
	}
	return False;
}

/* 
 * Check existing events in queue to find if any match.  If so, return.
 * If not, flush buffer and see if any more events are readable. If one
 * matches, return.  If all else fails, tell the user no events found.
 */
Bool XCheckTypedEventMQ (dpy, event_queue, type, event)
	Display* dpy;
        register XQueue *event_queue;
	register int type;	/* Selected event type. */
	XEvent *event;		/* XEvent to be filled in. */
{
	register _XQEvent *prev, *qelt;
	int n;			/* time through count */

	prev = NULL;
	for (n = 3; --n >= 0;) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		 qelt;
		 prev = qelt, qelt = qelt->next) {
		if (qelt->event.type == type) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }
	    UnlockMutex(&event_queue->lock);
	    switch (n) {
	      case 2:
		_XEventsQueued(dpy, QueuedAfterReading);
		break;
	      case 1:
		XFlush(dpy);
		break;
	    }
	}
	return False;
}

/* 
 * Check existing events in queue to find if any match.  If so, return.
 * If not, flush buffer and see if any more events are readable. If one
 * matches, return.  If all else fails, tell the user no events found.
 */
Bool XCheckTypedWindowEventMQ (dpy, event_queue, w, type, event)
	Display* dpy;
        register XQueue *event_queue;
	register Window w;	/* Selected window. */
	register int type;	/* Selected event type. */
	XEvent *event;		/* XEvent to be filled in. */
{
	register _XQEvent *prev, *qelt;
	int n;			/* time through count */

	prev = NULL;
	for (n = 3; --n >= 0;) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		 qelt;
		 prev = qelt, qelt = qelt->next) {
		if ((qelt->event.xany.window == w) &&
		    (qelt->event.type == type)) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }
	    UnlockMutex(&event_queue->lock);
	    switch (n) {
	      case 2:
		_XEventsQueued(dpy, event_queue, QueuedAfterReading);
		break;
	      case 1:
		XFlush(dpy);
		break;
	    }
	}
	return False;
}

/* 
 * Check existing events in queue to find if any match.  If so, return.
 * If not, flush buffer and see if any more events are readable. If one
 * matches, return.  If all else fails, tell the user no events found.
 */
Bool XCheckWindowEventMQ (dpy, event_queue, w, mask, event)
	Display* dpy;
        register XQueue *event_queue;
	register Window w;	/* Selected window. */
	register long mask;	/* Selected event mask. */
	XEvent *event;		/* XEvent to be filled in. */
{
 	register _XQEvent *prev, *qelt;
	int n;			/* time through count */

	prev = NULL;
	for (n = 3; --n >= 0;) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		 qelt;
		 prev = qelt, qelt = qelt->next) {
		if ((qelt->event.xany.window == w) &&
		    (qelt->event.type < LASTEvent) &&
		    (_Xevent_to_mask[qelt->event.type] & mask) &&
		    ((qelt->event.type != MotionNotify) ||
		     (mask & AllPointers) ||
		     (mask & AllButtons & qelt->event.xmotion.state))) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }
	    UnlockMutex(&event_queue->lock);
	    switch (n) {
	      case 2:
		_XEventsQueued(dpy, QueuedAfterReading);
		break;
	      case 1:
		XFlush(dpy);
		break;
	    }
	}
	return False;
}

/*
 * Read in pending events if needed and return the number of queued events.
 */
int XEventsQueuedMQ (dpy, event_queue, mode)
	Display* dpy;
        register XQueue *event_queue;
	register int mode;
{
        register int ret_val;

	LockMutex(&event_queue->lock);
	ret_val = event_queue->qlen;
	UnlockMutex(&event_queue->lock);

	if (ret_val || (mode == QueuedAlready))
	    return ret_val;
	else
	    return _XEventsQueued (dpy, event_queue, mode);
}

/* 
 * Flush output and (wait for and) return the next event matching the
 * predicate in the queue.
 */
XIfEventMQ (dpy, event_queue, event, predicate, arg)
    Display* dpy;
    register XQueue *event_queue;
    XEvent *event;
    XPredicate predicate;		/* function to call */
    XPointer *arg;
{
    register _XQEvent *qelt, *prev;
    int my_qlen;

    prev = NULL;
    while (True) {
	LockMutex(&event_queue->lock);
	for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
	    qelt;
	    prev = qelt, qelt = qelt->next) {
	    if ((*predicate)(dpy, &qelt->event, arg)) {
		*event = qelt->event;
		if (prev) {
		    if ((prev->next = qelt->next) == NULL)
			event_queue->tail = (XPointer)prev;
		} else {
		    if ((event_queue->head = (XPointer)qelt->next) == NULL)
		    event_queue->tail = NULL;
		}
		event_queue->qlen--;
		UnlockMutex(&event_queue->lock);
		_XFreeBlock(qelt);
		return True;
	    }
	}
	my_qlen = event_queue->qlen;
	UnlockMutex(&event_queue->lock);
	_XReadEvents(dpy, event_queue, my_qlen, NULL);
    }
}

/* 
 * Flush output and (wait for and) return the next event matching the
 * predicate in the queue. Return false if no match within timeout specified.
 */
Bool XIfEventTimeoutMQ (dpy, event_queue, event, predicate, arg, timeout)
	Display* dpy;
        register XQueue *event_queue;
	XEvent *event;
	XPredicate predicate;		/* function to call */
	XPointer *arg;
	Timeout timeout;		/* in milliseconds */
{
	register _XQEvent *qelt, *prev;
	int my_qlen;
	struct timeval time_to_return;
	Bool timed_out = False;

	time_to_return.tv_sec = 0;
	time_to_return.tv_usec = 0;
	if (timeout != WAITFOREVER && timeout != NOWAIT) {
	    gettimeofday(&time_to_return, NULL);
	    time_to_return.tv_usec += (timeout % 1000) * 1000;
	    time_to_return.tv_sec += timeout / 1000
				  + time_to_return.tv_usec / 1000000;
	    time_to_return.tv_usec %= 1000000;
	}	/* if (timeout != WAITFOREVER */

	prev = NULL;
	while (True) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((*predicate)(dpy, &qelt->event, arg)) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }	/* for */
	    my_qlen = event_queue->qlen;
	    UnlockMutex(&event_queue->lock);

	    if (timed_out)
		return False;

	    /* status 0 means _XReadEvents has timed out */
	    timed_out = ! _XReadEvents(dpy, event_queue, my_qlen,
			timeout == WAITFOREVER ? NULL : &time_to_return);
	}	/* while (True) */
}

/* 
 * Return the next event in the queue matching one of the events in the mask.
 * If no event, flush output, and wait until match succeeds.
 * Events earlier in the queue are not discarded.
 */
XMaskEventMQ (dpy, event_queue, mask, event)
	Display* dpy;
        register XQueue *event_queue;
	register long mask;		/* Selected event mask. */
	XEvent *event;			/* XEvent to be filled in. */
{
	register _XQEvent *prev, *qelt;
	int	my_qlen;

	prev = NULL;
	while (True) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((_Xevent_to_mask[qelt->event.type] & mask) &&
		    ((qelt->event.type != MotionNotify) ||
		     (mask & AllPointers) ||
		     (mask & AllButtons & qelt->event.xmotion.state))) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }
	    my_qlen = event_queue->qlen;
	    UnlockMutex(&event_queue->lock);
	    _XReadEvents(dpy, event_queue, my_qlen, NULL);
	}
}

/* 
 * Return the next event in the queue matching one of the events in the mask.
 * If no event, flush output, and wait until match succeeds or timeout.
 * Events earlier in the queue are not discarded. 
 */
Bool XMaskEventTimeoutMQ (dpy, event_queue, mask, event, timeout)
	Display* dpy;
        register XQueue *event_queue;
	register long mask;		/* Selected event mask. */
	XEvent *event;			/* XEvent to be filled in. */
	Timeout timeout;		/* in milliseconds */
{
	register _XQEvent *prev, *qelt;
	int my_qlen;
	struct timeval time_to_return;
	Bool timed_out = False;

	time_to_return.tv_sec = 0;
	time_to_return.tv_usec = 0;
	if (timeout != WAITFOREVER && timeout != NOWAIT) {
	    gettimeofday(&time_to_return, NULL);
	    time_to_return.tv_usec += (timeout % 1000) * 1000;
	    time_to_return.tv_sec += timeout / 1000
				  + time_to_return.tv_usec / 1000000;
	    time_to_return.tv_usec %= 1000000;
	}	/* if (timeout != WAITFOREVER */

	prev = NULL;
	while (True) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((_Xevent_to_mask[qelt->event.type] & mask) &&
		    ((qelt->event.type != MotionNotify) ||
		     (mask & AllPointers) ||
		     (mask & AllButtons & qelt->event.xmotion.state))) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }	/* for */
	    my_qlen = event_queue->qlen;
	    UnlockMutex(&event_queue->lock);

	    if (timed_out)
		return False;

	    /* status 0 means _XReadEvents has timed out */
	    timed_out = ! _XReadEvents(dpy, event_queue, my_qlen,
			timeout == WAITFOREVER ? NULL : &time_to_return);
	}	/* while (True) */
}

/* 
 * Return next event in queue, or if none, flush output and wait for
 * events.
 */
XNextEventMQ (dpy, event_queue, event)
	Display* dpy;
        register XQueue *event_queue;
	register XEvent *event;
{
	register _XQEvent *qelt;

	LockMutex(&event_queue->lock);
	if (event_queue->qlen == 0) {
	    UnlockMutex(&event_queue->lock);
	    _XReadEvents(dpy, event_queue, 0, NULL);
	    LockMutex(&event_queue->lock);
	}
	qelt = (_XQEvent *)event_queue->head;
	*event = qelt->event;

	/* move the head of the queue to the free list */
	if ((event_queue->head = (XPointer)qelt->next) == NULL)
	    event_queue->tail = NULL;
	event_queue->qlen--;
	UnlockMutex(&event_queue->lock);
	_XFreeBlock(qelt);
}

/* 
 * Return next event in queue, or if none, flush output and wait for
 * events until timeout. If timed out, tell the user no event returned.
 */
Bool XNextEventTimeoutMQ (dpy, event_queue, event, timeout)
	Display* dpy;
        register XQueue *event_queue;
	XEvent *event;
	register Timeout timeout;	/* in milliseconds */
{
	register _XQEvent *qelt;
	struct timeval time_to_return;

	time_to_return.tv_sec = 0;
	time_to_return.tv_usec = 0;
	if (timeout != WAITFOREVER && timeout != NOWAIT) {
	    gettimeofday(&time_to_return, NULL);
	    time_to_return.tv_usec += (timeout % 1000) * 1000;
	    time_to_return.tv_sec += timeout / 1000
				  + time_to_return.tv_usec / 1000000;
	    time_to_return.tv_usec %= 1000000;
	}	/* if (timeout != WAITFOREVER */

	LockMutex(&event_queue->lock);
	if (event_queue->qlen == 0) {
	    UnlockMutex(&event_queue->lock);
	    if (! _XReadEvents(dpy, event_queue, 0,
		timeout == WAITFOREVER ? NULL : &time_to_return))
		return False;	/* status 0 means _XReadEvents has timed out */
	    else
		LockMutex(&event_queue->lock);
	}
	qelt = (_XQEvent *)event_queue->head;
	*event = qelt->event;

	/* move the head of the queue to the free list */
	if ((event_queue->head = (XPointer)qelt->next) == NULL)
	    event_queue->tail = NULL;
	event_queue->qlen--;
	UnlockMutex(&event_queue->lock);
	_XFreeBlock(qelt);
	return True;
}

/* 
 * Return the next event in the queue, BUT do not remove it from the queue.
 * If none found, flush and wait until there is an event to peek.
 */
XPeekEventMQ (dpy, event_queue, event)
	Display* dpy;
        register XQueue *event_queue;
	register XEvent *event;
{
	LockMutex(&event_queue->lock);
	if (event_queue->qlen == 0) {
	    UnlockMutex(&event_queue->lock);
	    _XReadEvents(dpy, event_queue, 0, NULL);
	    LockMutex(&event_queue->lock);
	}
	*event = ((_XQEvent *)event_queue->head)->event;
	UnlockMutex(&event_queue->lock);
}

/* 
 * Return the next event in the queue, BUT do not remove it from the queue.
 * If none found, flush and wait until timeout. If timed out, tell the user
 * no event returned.
 */
Bool XPeekEventTimeoutMQ (dpy, event_queue, event, timeout)
	Display* dpy;
        register XQueue *event_queue;
	register XEvent *event;
	register Timeout timeout;	/* in milliseconds */
{
	struct timeval time_to_return;

	time_to_return.tv_usec = 0;
	time_to_return.tv_sec = 0;
	if (timeout != WAITFOREVER && timeout != NOWAIT) {
	    gettimeofday(&time_to_return, NULL);
	    time_to_return.tv_usec += (timeout % 1000) * 1000;
	    time_to_return.tv_sec += timeout / 1000
				  + time_to_return.tv_usec / 1000000;
	    time_to_return.tv_usec %= 1000000;
	}	/* if (timeout != WAITFOREVER */

	LockMutex(&event_queue->lock);
	if (event_queue->qlen == 0) {
	    UnlockMutex(&event_queue->lock);
	    if (! _XReadEvents(dpy, event_queue, 0,
		timeout == WAITFOREVER ? NULL : &time_to_return))
		return False;	/* status 0 means _XReadEvents has timed out */
	    else
		LockMutex(&event_queue->lock);
	}
	*event = ((_XQEvent *)event_queue->head)->event;
	UnlockMutex(&event_queue->lock);
	return True;
}

/*
 * Return the next event in the queue that satisfies the predicate.
 * BUT do not remove it from the queue.
 * If none found, flush, and then wait until one satisfies the predicate.
 */
XPeekIfEventMQ (dpy, event_queue, event, predicate, arg)
	Display* dpy;
        register XQueue *event_queue;
	XEvent *event;
	register XPredicate predicate;
	XPointer *arg;
{
	register _XQEvent *qelt, *prev;
	int my_qlen;

	prev = NULL;
	while (True) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((*predicate)(dpy, &qelt->event, arg)) {
		    *event = qelt->event;
		    UnlockMutex(&event_queue->lock);
		    return;
		}
	    }
	    my_qlen = event_queue->qlen;
	    UnlockMutex(&event_queue->lock);
	    _XReadEvents(dpy, event_queue, my_qlen, NULL);
	}
}

/*
 * Return the next event in the queue that satisfies the predicate.
 * BUT do not remove it from the queue.
 * If none found, flush, and then wait until timeout or one satisfies
 * the predicate.
 */
Bool XPeekIfEventTimeoutMQ (dpy, event_queue, event, predicate, arg, timeout)
	Display* dpy;
        register XQueue *event_queue;
	XEvent *event;
	register XPredicate predicate;		/* function to call */
	register XPointer *arg;
	Timeout timeout;			/* in milliseconds */
{
	register _XQEvent *qelt, *prev;
	int my_qlen;
	struct timeval time_to_return;
	Bool timed_out = False;

	time_to_return.tv_usec = 0;
	time_to_return.tv_sec = 0;
	if (timeout != WAITFOREVER && timeout != NOWAIT) {
	    gettimeofday(&time_to_return, NULL);
	    time_to_return.tv_usec += (timeout % 1000) * 1000;
	    time_to_return.tv_sec += timeout / 1000
				  + time_to_return.tv_usec / 1000000;
	    time_to_return.tv_usec %= 1000000;
	}	/* if (timeout != WAITFOREVER */

	prev = NULL;
	while (True) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((*predicate)(dpy, &qelt->event, arg)) {
		    *event = qelt->event;
		    UnlockMutex(&event_queue->lock);
		    return True;
		}
	    }	/* for */
	    my_qlen = event_queue->qlen;
	    UnlockMutex(&event_queue->lock);

	    if (timed_out)
		return False;

	    /* status 0 means _XReadEvents has timed out */
	    timed_out = ! _XReadEvents(dpy, event_queue, my_qlen,
			timeout == WAITFOREVER ? NULL : &time_to_return);
	}	/* while (True) */
}

/*
 * Read in pending events if needed and return the number of queued events.
 */
int XPendingMQ (dpy, event_queue)
	Display* dpy;
	register XQueue *event_queue;
{
	register int ret_val;

	LockMutex(&event_queue->lock);
	ret_val = event_queue->qlen;
	UnlockMutex(&event_queue->lock);

	if (ret_val)
	    return ret_val;
	else
	    return _XEventsQueued (dpy, event_queue, QueuedAfterFlush);
}

/*
 * Put an event back at the head of the queue.
 */
XPutBackEventMQ (dpy, event_queue, event)
	Display* dpy;
        register XQueue *event_queue;
	XEvent *event;
{
	register _XQEvent *qelt;

	qelt = (_XQEvent *)_XGetBlock();
	qelt->event = *event;
	LockMutex(&event_queue->lock);
	qelt->next = (_XQEvent *)event_queue->head;	/* to the head of queue */
	event_queue->head = (XPointer)qelt;
	if (event_queue->tail == NULL)
	    event_queue->tail = (XPointer)qelt;
	event_queue->qlen++;
	UnlockMutex(&event_queue->lock);
}

/*
 * Return the number of events in the queue
 */
int XQLengthMQ (dpy, event_queue)
	Display* dpy;
        register XQueue *event_queue;
{
	int ret_val;

	LockMutex(&event_queue->lock);
	ret_val = event_queue->qlen;
        UnlockMutex(&event_queue->lock);
	return ret_val;
}

/*
 * Synchronize with errors and events, optionally discarding pending events
 * in the specified event queue.
 */
XSyncMQ (dpy, event_queue, discard_events)
	register Display* dpy;
        register XQueue *event_queue;
	register Bool discard_events;
{
	_XQReply *qreply;
	_XQReply *rid;
	xGetInputFocusReply *rep;
	register xReq *req;

	LockDisplay(dpy);
	GetEmptyReq(GetInputFocus, req);
	rid = _XInitReply(dpy, False, SIZEOF(xGetInputFocusReply));
	UnlockDisplay(dpy);

	qreply = _XReply(dpy, rid);
	FreeQReply(qreply);

    	if (discard_events) {
    	    LockMutex(&event_queue->lock);
	    _XFreeEvents(event_queue);
	    UnlockMutex(&event_queue->lock);
	}
}

/* 
 * Return the next event in the queue
 * for the given window matching one of the events in the mask.
 * Events earlier in the queue are not discarded.
 * If none found, flush, and then wait until an event arrives which
 * matches.
 */
XWindowEventMQ (dpy, event_queue, w, mask, event)
	Display* dpy;
        register XQueue *event_queue;
	register Window w;		/* Selected window. */
	register long mask;		/* Selected event mask. */
	XEvent *event;	/* XEvent to be filled in. */
{
	register _XQEvent *qelt, *prev;
	int	my_qlen;

	prev = NULL;
	while (True) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((qelt->event.xany.window == w) &&
		    (_Xevent_to_mask[qelt->event.type] & mask) &&
		    ((qelt->event.type != MotionNotify) ||
		     (mask & AllPointers) || (mask & AllButtons &
		     qelt->event.xmotion.state))) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }
	    my_qlen = event_queue->qlen;
	    UnlockMutex(&event_queue->lock);
	    _XReadEvents(dpy, event_queue, my_qlen, NULL);
	}
}

/* 
 * Return the next event in the queue
 * for the given window matching one of the events in the mask.
 * Events earlier in the queue are not discarded.
 * If none found, flush, and then wait until timeout.
 * If timed out and still none matches, tell the user no event returned.
 */
Bool XWindowEventTimeoutMQ (dpy, event_queue, w, mask, event, timeout)
	Display* dpy;
        register XQueue *event_queue;
	register Window w;		/* Selected window. */
	register long mask;		/* Selected event mask. */
	XEvent *event;			/* XEvent to be filled in. */
	Timeout timeout;		/* in milliseconds */
{
	register _XQEvent *qelt, *prev;
	int my_qlen;
	struct timeval time_to_return;
	Bool timed_out = False;

	time_to_return.tv_usec = 0;
	time_to_return.tv_sec = 0;
	if (timeout != WAITFOREVER && timeout != NOWAIT) {
	    gettimeofday(&time_to_return, NULL);
	    time_to_return.tv_usec += (timeout % 1000) * 1000;
	    time_to_return.tv_sec += timeout / 1000
				  + time_to_return.tv_usec / 1000000;
	    time_to_return.tv_usec %= 1000000;
	}	/* if (timeout != WAITFOREVER */

	prev = NULL;
	while (True) {
	    LockMutex(&event_queue->lock);
	    for (qelt = prev ? prev->next : (_XQEvent *)event_queue->head;
		qelt;
		prev = qelt, qelt = qelt->next) {
		if ((qelt->event.xany.window == w) &&
		    (_Xevent_to_mask[qelt->event.type] & mask) &&
		    ((qelt->event.type != MotionNotify) ||
		     (mask & AllPointers) || (mask & AllButtons &
		     qelt->event.xmotion.state))) {
		    *event = qelt->event;
		    if (prev) {
			if ((prev->next = qelt->next) == NULL)
			    event_queue->tail = (XPointer)prev;
		    } else {
			if ((event_queue->head = (XPointer)qelt->next) == NULL)
			event_queue->tail = NULL;
		    }
		    event_queue->qlen--;
		    UnlockMutex(&event_queue->lock);
		    _XFreeBlock(qelt);
		    return True;
		}
	    }	/* for */
	    my_qlen = event_queue->qlen;
	    UnlockMutex(&event_queue->lock);

	    if (timed_out)
		return False;

	    /* status 0 means _XReadEvents has timed out */
	    timed_out = ! _XReadEvents(dpy, event_queue, my_qlen,
			timeout == WAITFOREVER ? NULL : &time_to_return);
	}	/* while (True) */
}

#endif	/* MULTI_THREAD */
