/* $XConsortium: XGetFPath.c,v 11.15 91/01/06 11:45:55 rws Exp $ */
/* Copyright    Massachusetts Institute of Technology    1986	*/

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

#define NEED_REPLIES
#include "Xlibint.h"

char **XGetFontPath(dpy, npaths)
register Display *dpy;
int *npaths;	/* RETURN */
{
#ifdef MULTI_THREAD
    _XQReply *qreply, *rid;
    xGetFontPathReply *rep;
#else
    xGetFontPathReply rep;
#endif
    register long nbytes;
    char **flist;
    char *ch;
    register unsigned i;
    register int length;
    register xReq *req;

    LockDisplay(dpy);
    GetEmptyReq (GetFontPath, req);

#ifdef MULTI_THREAD
    rid = _XInitReply( dpy, False, SIZEOF(xGetFontPathReply) );

    UnlockDisplay( dpy );

    qreply = _XReply( dpy, rid );

    rep = (xGetFontPathReply *)qreply->reply;

    if (rep->nPaths) {
	flist = (char **)
	    Xmalloc((unsigned) rep->nPaths * sizeof (char *));

	/* Figure out how long all those strings really are. */
	ch = qreply->data;
	for ( i = 0; i < rep->nPaths; i++)
	    ch += *ch;

	/* When we get here, ch points just past the last item */

	nbytes = ch - qreply->data;

	/* Steal the data block from qreply, if possible
	 *
	 * To be stealable, it has to have padding, and it has to be from
	 *  Xmalloc.  Note that this depends on qreply->data being reasonably
	 *  aligned, which depends on the 32bitness of X connections.
	 */

	if ( (qreply->data_size > sizeof( _XQBlock )) && (nbytes & 3) ) {
	    /* I can steal it! */
	    ch = qreply->data;
	    qreply->data = NULL;
	    qreply->data_size = 0;
	}
	else {
	    ch = (char *) Xmalloc ((unsigned) (nbytes + 1));
		/* +1 to leave room for last null-terminator */
	}

	if ((! flist) || (! ch)) {
	    if (flist) Xfree((char *) flist);
	    if (ch) Xfree(ch);
	    FreeQReply( qreply );
	    SyncHandle();
	    return (char **) NULL;
	}

	if (qreply->data)	/* stolen data block? */
	    bcopy( qreply->data, ch, nbytes );	/* Guess not. */

	/*
	 * unpack into null terminated strings.
	 */
	length = *ch;
	for (i = 0; i < rep->nPaths; i++) {
	    flist[i] = ch+1;  /* skip over length */
	    ch += length + 1; /* find next length ... */
	    length = *ch;
	    *ch = '\0'; /* and replace with null-termination */
	}
    }
    else flist = NULL;
    *npaths = rep->nPaths;

    FreeQReply( qreply );
#else	/* MULTI_THREAD */
    (void) _XReply (dpy, (xReply *) &rep, 0, xFalse);

    if (rep.nPaths) {
	flist = (char **)
	    Xmalloc((unsigned) rep.nPaths * sizeof (char *));
	nbytes = (long)rep.length << 2;
	ch = (char *) Xmalloc ((unsigned) (nbytes + 1));
	    /* +1 to leave room for last null-terminator */

	if ((! flist) || (! ch)) {
	    if (flist) Xfree((char *) flist);
	    if (ch) Xfree(ch);
	    _XEatData(dpy, (unsigned long) nbytes);
	    UnlockDisplay(dpy);
	    SyncHandle();
	    return (char **) NULL;
	}

	_XReadPad (dpy, ch, nbytes);
	/*
	 * unpack into null terminated strings.
	 */
	length = *ch;
	for (i = 0; i < rep.nPaths; i++) {
	    flist[i] = ch+1;  /* skip over length */
	    ch += length + 1; /* find next length ... */
	    length = *ch;
	    *ch = '\0'; /* and replace with null-termination */
	}
    }
    else flist = NULL;
    *npaths = rep.nPaths;
    UnlockDisplay(dpy);
#endif	/* MULTI_THREAD */
    SyncHandle();
    return (flist);
}

XFreeFontPath (list)
char **list;
{
    if (list != NULL) {
	Xfree (list[0]-1);
	Xfree ((char *)list);
    }
}
