/* $XConsortium: XGetPntMap.c,v 1.14 91/01/06 11:46:14 rws Exp $ */
/* Copyright    Massachusetts Institute of Technology    1986	*/

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

#define NEED_REPLIES
#include "Xlibint.h"

#ifdef MIN		/* some systems define this in <sys/param.h> */
#undef MIN
#endif
#define MIN(a, b) ((a) < (b) ? (a) : (b))

int XGetPointerMapping (dpy, map, nmaps)
    register Display *dpy;
    unsigned char *map;	/* RETURN */
    int nmaps;

{
    long nbytes;
    register xReq *req;
#ifdef MULTI_THREAD
    _XQReply *qreply;
    _XQReply *rid;
    xGetPointerMappingReply *rep;
    int nElts;
#else
    xGetPointerMappingReply rep;
    unsigned char mapping[256];	/* known fixed size */
#endif

    LockDisplay(dpy);
    GetEmptyReq(GetPointerMapping, req);
#ifdef MULTI_THREAD
    rid = _XInitReply( dpy, False, SIZEOF(xGetPointerMappingReply) );

    UnlockDisplay( dpy );

    qreply = _XReply( dpy, rid );

    rep = (xGetPointerMappingReply *)qreply->reply;

    nbytes = (long)rep->length << 2;
    if (rep->nElts)
	bcopy( (char *) qreply->data, (char *) map,
		MIN((int)rep->nElts, nmaps) );

    nElts = rep->nElts;
    FreeQReply( qreply );

    SyncHandle();

    return nElts;
#else
    (void) _XReply(dpy, (xReply *)&rep, 0, xFalse);

    nbytes = (long)rep.length << 2;
    _XRead (dpy, (char *)mapping, nbytes);
    /* don't return more data than the user asked for. */
    if (rep.nElts) {
	    bcopy ((char *) mapping, (char *) map, 
		MIN((int)rep.nElts, nmaps) );
	}
    UnlockDisplay(dpy);
    SyncHandle();
    return ((int) rep.nElts);
#endif	/* MULTI_THREAD */
}

#if NeedFunctionPrototypes
KeySym *XGetKeyboardMapping (Display *dpy,
#if NeedWidePrototypes
			     unsigned int first_keycode,
#else
			     KeyCode first_keycode,
#endif
			     int count,
			     int *keysyms_per_keycode)
#else
KeySym *XGetKeyboardMapping (dpy, first_keycode, count, keysyms_per_keycode)
    register Display *dpy;
    KeyCode first_keycode;
    int count;
    int *keysyms_per_keycode;		/* RETURN */
#endif
{
    long nbytes;
    unsigned long nkeysyms;
    register KeySym *mapping = NULL;
    register xGetKeyboardMappingReq *req;
#ifdef MULTI_THREAD
    _XQReply *qreply;
    _XQReply *rid;
    xGetKeyboardMappingReply *rep;
#else
    xGetKeyboardMappingReply rep;
#endif

    LockDisplay(dpy);
    GetReq(GetKeyboardMapping, req);
    req->firstKeyCode = first_keycode;
    req->count = count;
#ifdef MULTI_THREAD
    rid = _XInitReply( dpy, False, SIZEOF(xGetKeyboardMappingReply) );

    UnlockDisplay( dpy );

    qreply = _XReply( dpy, rid );
    if (qreply->status == 0) {
	FreeQReply( qreply );
	SyncHandle();
	return (KeySym *) NULL;
    }

    rep = (xGetKeyboardMappingReply *)qreply->reply;

    nkeysyms = (unsigned long) rep->length;

    if (nkeysyms > 0) {
	nbytes = nkeysyms * sizeof (KeySym);
	mapping = (KeySym *) Xmalloc ((unsigned) nbytes);
	nbytes = nkeysyms << 2;
	if (! mapping) {
	    FreeQReply( qreply );
	    SyncHandle();
	    return (KeySym *) NULL;
	}
	bcopy((char *)qreply->data, (char *)mapping, nbytes);
    }
    *keysyms_per_keycode = rep->keySymsPerKeyCode;

    FreeQReply( qreply );
#else
    if (! _XReply(dpy, (xReply *)&rep, 0, xFalse)) {
	UnlockDisplay(dpy);
	SyncHandle();
	return (KeySym *) NULL;
    }

    nkeysyms = (unsigned long) rep.length;
    if (nkeysyms > 0) {
	nbytes = nkeysyms * sizeof (KeySym);
	mapping = (KeySym *) Xmalloc ((unsigned) nbytes);
	nbytes = nkeysyms << 2;
	if (! mapping) {
	    _XEatData(dpy, (unsigned long) nbytes);
	    UnlockDisplay(dpy);
	    SyncHandle();
	    return (KeySym *) NULL;
	}
	_XRead32 (dpy, (char *) mapping, nbytes);
    }
    *keysyms_per_keycode = rep.keySymsPerKeyCode;
    UnlockDisplay(dpy);
#endif	/* MULTI_THREAD */
    SyncHandle();
    return (mapping);
}

