/* $XConsortium: XListExt.c,v 11.10 91/01/06 11:46:50 rws Exp $ */
/* Copyright    Massachusetts Institute of Technology    1986	*/

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

#define NEED_REPLIES
#include "Xlibint.h"

char **XListExtensions(dpy, nextensions)
register Display *dpy;
int *nextensions;	/* RETURN */
{
	char **list;
	char *ch;
	register unsigned i;
	register int length;
	register xReq *req;
	register long rlen;
#ifdef MULTI_THREAD
	_XQReply *qreply, *rid;
	xListExtensionsReply *rep;
#else
	xListExtensionsReply rep;
#endif

	LockDisplay(dpy);
	GetEmptyReq (ListExtensions, req);

#ifdef MULTI_THREAD
	rid = _XInitReply( dpy, False, SIZEOF(xListExtensionsReply) );

	UnlockDisplay( dpy );

	qreply = _XReply( dpy, rid );

	if (!qreply->status) {
	    FreeQReply( qreply );
	    SyncHandle();
	    return (char **) NULL;
	}

	rep = (xListExtensionsReply *)qreply->reply;

	if (rep->nExtensions) {
	    list = (char **) Xmalloc(
		(unsigned)(rep->nExtensions * sizeof(char *)));

	    ch = qreply->data;
	    for (i = 0; i < rep->nExtensions; i++)
		ch += *ch;

	    /* When we get here, ch points just past the last item */
	    rlen = ch - qreply->data;

	    if ( (qreply->data_size > sizeof( _XQBlock )) && (rlen & 3) ) {
		/* I can steal it! */
		ch = qreply->data;
		qreply->data = NULL;
		qreply->data_size = 0;
	    }
	    else {
		ch = (char *) Xmalloc ((unsigned) rlen + 1);
		    /* +1 to leave room for last null-terminator */
	    }	    

	    if (!list || !ch) {
		if (list) Xfree((char *) list);
		if (ch)   Xfree((char *) ch);
		FreeQReply( qreply );
		SyncHandle();
		return (char **) NULL;
	    }

	    if (qreply->data)	/* did I steal it? */
		bcopy( qreply->data, ch, rlen );	/* Guess not. */

	    /*
	     * unpack into null terminated strings.
	     */
	    length = *ch;
	    for (i = 0; i < rep->nExtensions; i++) {
		list[i] = ch+1;  /* skip over length */
		ch += length + 1; /* find next length ... */
		length = *ch;
		*ch = '\0'; /* and replace with null-termination */
	    }
	}
	else list = (char **) NULL;

	*nextensions = rep->nExtensions;

	FreeQReply( qreply );
#else
	if (! _XReply (dpy, (xReply *) &rep, 0, xFalse)) {
	    UnlockDisplay(dpy);
	    SyncHandle();
	    return (char **) NULL;
	}

	if (rep.nExtensions) {
	    list = (char **) Xmalloc (
                (unsigned)(rep.nExtensions * sizeof (char *)));
	    rlen = rep.length << 2;
	    ch = (char *) Xmalloc ((unsigned) rlen + 1);
                /* +1 to leave room for last null-terminator */

	    if ((!list) || (!ch)) {
		if (list) Xfree((char *) list);
		if (ch)   Xfree((char *) ch);
		_XEatData(dpy, (unsigned long) rlen);
		UnlockDisplay(dpy);
		SyncHandle();
		return (char **) NULL;
	    }

	    _XReadPad (dpy, ch, rlen);
	    /*
	     * unpack into null terminated strings.
	     */
	    length = *ch;
	    for (i = 0; i < rep.nExtensions; i++) {
		list[i] = ch+1;  /* skip over length */
		ch += length + 1; /* find next length ... */
		length = *ch;
		*ch = '\0'; /* and replace with null-termination */
	    }
	}
	else list = (char **) NULL;

	*nextensions = rep.nExtensions;
	UnlockDisplay(dpy);

#endif	/* MULTI_THREAD */

	SyncHandle();
	return (list);
}

XFreeExtensionList (list)
char **list;
{
	if (list != NULL) {
	    Xfree (list[0]-1);
	    Xfree ((char *)list);
	}
}
