/* Copyright	Xerox Corporation	1991, 1992
 *
 * History
 * Tue Dec 31 10:18:35 PST 1991	SEW	Created
 *
 */

/* Copyright (C) 1992 by Digital Equipment Corporation and Xerox Corporation */

/*	All Rights Reserved
Permission to use, copy, modify, and distribute this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Digital and Xerox not
be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  Digital and
Xerox make no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.
*/

/*
 * XPollDisplays - Perform a poll() function on the specified list of
 *   displays, file descriptors, and the timeout.
 *
 * Poll was chosen instead of select() because the poll() system call
 *   can *always* be emulated by select, but select may not be able to
 *   emulate poll() (on systems that support many more file handles than
 *   usual, for example).
 *
 * The hope is that toolkits will use XPollDisplays rather than reaching
 *   into the display structure (with XConnectionNumber), thus avoiding
 *   another source of OS-dependencies in the toolkit.
 */

typedef struct
{
    Display *dpy;
    short events;
    short revents;
} PollDisplay;

#include "Xlibint.h"

#if NeedFunctionPrototypes
int XPollDisplays(
    PollDisplay *	dpyList,
    int			ndpys,
    struct pollfd *	fds,
    int			nfds,
    int			timeout	)
#else
Status XPollDisplays( dpyList, ndpys, fds, nfds, timeout )
    PollDisplay *	dpyList;
    unsigned long	ndpys;
    struct pollfd *	fds;
    unsigned long	nfds;
    int			timeout;
#endif
{
    struct pollfd *fdlist;
    int i, fdlCount;
    int ret;

    /* Check for easy case */
    if (ndpys == 0 && nfds == 0 && timeout == 0)
	return 0;

    if (ndpys == 0)	/* Just call poll() */
	return poll( fds, nfds, timeout );

    fdlist = Xmalloc( (ndpys + nfds) * sizeof(struct pollfd) );
    if (!fdlist)
    {
	errno = ENOMEM;
	return 0;
    }

    fdlCount = 0;

    for (i = 0; i < ndpys; i++)
    {
	fdlist[fdlCount].fd = dpyList[i].dpy->fd;
	fdlist[fdlCount].events = dpyList[i].events;
	fdlCount++;
    }

    bcopy( fds, &fdlist[fdlCount], nfds * sizeof(struct pollfd) );

    ret = poll( fdlist, ndpys + nfds, timeout );

    fdlCount = 0;
    for (i = 0; i < ndpys; i++)
	dpyList[i].revents = fdlist[fdlCount++].revents;

    for (i = 0; i < nfds; i++)
	fds[i].revents = fdlist[fdlCount++].revents;

    return ret;
}
