/*
 * ImageTest - test of get and put image with the MT Xlib
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	"XlibTK.h"
#include	"TestTK.h"


typedef struct _image_data
{
    Display	*dpy;
    Window	window;
    XImage	*image;
    Boolean	valid;
    Geometry	geometry;
    Pixmap	pixmap;
    GC		gc;
    Defaults	defaults;

}	*ImageData;


/*
 * local routines
 */

static void GetImage(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    Window	/* window */,
    XEvent *	/* event */,
    void *	/* data */
#endif
);
static void PutImage(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    Window	/* window */,
    XEvent *	/* event */,
    void *	/* data */
#endif
);
static void Cleanup(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    void *	/* data */
#endif
);

/*
 * start of InitImageTest function
 */

void
InitImageTest( context, test_data )
    AppContext	context;
    TestData	test_data;
{
    int		depth;
    ImageData	image_data;

    image_data = (ImageData)malloc( sizeof( *image_data ));
    image_data->defaults = test_data->defaults;
    image_data->dpy = TKDpyFromContext( context );
    image_data->window = test_data->window;
    image_data->valid = FALSE;
    image_data->gc = test_data->defaults->normGC;
    depth = DefaultDepthOfScreen( DefaultScreenOfDisplay( image_data->dpy ));
    image_data->pixmap = XCreatePixmap( image_data->dpy, image_data->window,
				       500, 500, depth );
/*
 * register event handlers
 */
    fprintf( stderr, "button press handler = %lx\n",
    TKAddEventHandler( context, GetImage, image_data, image_data->window,
		      ButtonPressMask ));
    fprintf( stderr, "button release handler = %lx\n",
    TKAddEventHandler( context, PutImage, image_data, image_data->window,
		      ButtonReleaseMask ));
    fprintf( stderr, "expose handler = %lx\n",
    TKAddEventHandler( context, PutImage, image_data, image_data->window,
		      ExposureMask ));
/*
 * register the cleanup proc
 */
    TKAddCleanupProc( context, Cleanup, image_data );
    return;

}	/* end of InitImageTest function */

/*
 * start of GetImage function
 */

static void
GetImage( context, window, event, data )
    AppContext		context;
    Window		window;
    XEvent		*event;
    void		*data;
{
    int		x;
    int		y;
    unsigned	width;
    unsigned	height;
    unsigned	depth;
    unsigned	border;
    Window	root;
    ImageData	image_data = (ImageData)data;

    fprintf( stderr, "GetImage: window = %d\n", window );
    if( ((XButtonEvent*)event)->state != None )
    {
	return;
    }

    if( XGetGeometry( image_data->dpy, window, &root, &x, &y, &width, &height,
		      &border, &depth ) == 0 )
    {
	fprintf( stderr, "XGetGeometry failed in GetImage\n" );
	exit( 1 );
    }

    width = MIN( width, 500 );
    height = MIN( height, 500 );
    image_data->geometry.width = width;
    image_data->geometry.height = height;

    if( image_data->valid )
    {
	free( image_data->image );
    }
    image_data->image = XGetImage( image_data->dpy, root, 0, 0, width, height,
				   ~0, XYPixmap );
/*
 * the XPutImage is done here so that I can test both XPutImage and XCopyArea
 */
    XPutImage( image_data->dpy, image_data->pixmap, image_data->gc,
	       image_data->image, 0, 0, 0, 0, width, height );
    image_data->valid = TRUE;

}	/* end of GetImage function */

/*
 * start of PutImage function
 */

static void
PutImage( context, window, event, data )
    AppContext		context;
    Window		window;
    XEvent		*event;
    void		*data;
{
    ImageData	image_data = (ImageData)data;

    fprintf( stderr, "PutImage: window = %d\n", window );
    if(( ! image_data->valid ) || ( event->xany.type == NoExpose ))
    {
	return;
    }

    XCopyArea( image_data->dpy, image_data->pixmap, image_data->window,
	      image_data->gc, 0, 0, image_data->geometry.width,
	      image_data->geometry.height, 0, 0 );

}	/* end of PutImage function */

/*
 * start of Cleanup function
 */

static void
Cleanup( context, data )
    AppContext	context;
    void	*data;
{
    fprintf( stderr, "Cleanup (ImageTest)\n" );
    free( ((ImageData)data)->image );
    XFreePixmap( ((ImageData)data)->dpy, ((ImageData)data)->pixmap );
    free( data );

}	/* end of Cleanup function */

