/*
 * TestTK.c - Toolkit for test programs
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<pthread.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/Xatom.h>
#include	"XlibTK.h"
#include	"TestTK.h"

/*
 * global data
 */

static pthread_mutex_t	TestLock;
static pthread_once_t	once = pthread_once_init;

struct _child
{
    struct _child	*next;
    pthread_t		thread;
};

typedef struct
{
    unsigned long	top_left_x;
    unsigned long	top_left_y;
    unsigned long	bot_left_x;
    unsigned long	bot_left_y;
    unsigned long	bot_right_x;
    unsigned long	bot_right_y;
    TestInfo		test_info;

}	*MenuList;
    
struct _run_data
{
    Defaults		defaults;
    struct _child	*children;
    AppContext		context;
    MenuList		menus;
    int			num_items;
};

/*
 * local routines
 */

static void InitLocks(
#if NeedFunctionPrototypes
    void
#endif
);
static void Cleanup(
#if NeedFunctionPrototypes
    struct _run_data*	/* run_data */
#endif
);
static void ClickHandler(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    Window	/* window */,
    XEvent *	/* *event */,
    void *	/* *data */
#endif
);
static void ExposeHandler(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    Window	/* window */,
    XEvent *	/* event */,
    void *	/* data */
#endif
);
static int
FindItem(
#if NeedFunctionPrototypes
    MenuList	/* menu */,
    int		/* num_items */,
    long int	/* x */,
    long int	/* y */
#endif
);
static void
DrawItem(
#if NeedFunctionPrototypes
    Display *	/* dpy */,
    Window	/* win */,
    GC		/* fg */,
    GC		/* bg */,
    MenuList	/* menu */
#endif
);
static void
DrawMenu(
#if NeedFunctionPrototypes
    Display *	/* dpy */,
    Window	/* win */,
    GC		/* fg */,
    GC		/* bg */,
    MenuList	/* menu */,
    int		/* num_items */
#endif
);
static void
ChildProc(
#if NeedFunctionPrototypes
    TestInfo	/* info needed to run the test */
#endif
);
static void CtlBtn1Handler(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    Window	/* window */,
    XEvent *	/* event */,
    void *	/* data */
#endif
);
static void DelPropHandler(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    Window	/* window */,
    XEvent *	/* event */,
    void *	/* data */
#endif
);

/*
 * start of InitLocks function
 */

static void
InitLocks()
{
    pthread_mutex_init( &TestLock, pthread_mutexattr_default );

}	/* end of InitLocks function */

/*
 * start of RunTest function
 */

void
RunTest( menu_def, num_items )
    TestInfo	menu_def;
    int		num_items;
{
    struct _run_data	*run_data;
    Defaults		defaults;
    MenuList		menu;
    Display		*dpy;
    AppContext		context;
    long int		width;
    long int		height;
    int			column = 0;
    unsigned		x = COLSPACE;
    unsigned		y = ROWSPACE;
    int			max_width = 0;
    int			max_ascent = 0;
    int			max_descent = 0;
    int			dir;
    int			ascent;
    int			descent;
    int			i;
    Window		top_win;
    GC			normGC;
    GC			revGC;
    XGCValues		gc_vals;
    XFontStruct 	*font_info;
    XSizeHints		size_hints;
    Atom		del_atom;
    XCharStruct		overall;
    unsigned long	black_pix;
    unsigned long	white_pix;
    HandlerList		click_handler;
    HandlerList		expose_handler;

    pthread_once( &once, InitLocks );
/*
 * initialize the Xlib toolkit
 */
    context = TKInitialize( NULL );
    if( context == NULL )
    {
	fprintf( stderr, "Failed initializing the toolkit.\n" );
	exit( 1 );
    }
    dpy = TKDpyFromContext( context );
/*
 * load a default font
 */
    font_info = XLoadQueryFont( dpy, FONTNAME );
    if( font_info == NULL )
    {
	fprintf( stderr, "Cannot open font matching - %s\n", FONTNAME );
	exit( EXIT_FAILURE );
    }
/*
 * create the menu array and determine the size for menu
 */
    menu = (MenuList)malloc( sizeof( *menu ) * num_items );
    for( i = 0; i < num_items; i++ )
    {
	menu[i].test_info = &menu_def[i];
	XQueryTextExtents( dpy, font_info->fid, menu[i].test_info->name,
			  strlen( menu[i].test_info->name ),
			  &dir, &ascent, &descent, &overall );
	menu[i].bot_left_x = overall.width / 2;

	max_width = MAX( max_width, overall.width );
	max_ascent = MAX( max_ascent, overall.ascent );
	max_descent = MAX( max_descent, overall.descent );
    }
    printf( "max_width = %d, max_ascent = %d, max_descent = %d\n",
	   max_width, max_ascent, max_descent );
/*
 * finish initializing the menu array
 */
    for( i = 0; i < num_items; i++ )
    {
	column++;
	menu[i].top_left_x = x;
	menu[i].top_left_y = y;
	menu[i].bot_left_x = x + (( max_width / 2 ) - menu[i].bot_left_x );
	menu[i].bot_left_y = y + max_ascent;
	menu[i].bot_right_x = x + max_width;
	menu[i].bot_right_y = y + max_ascent;

	if( column < NUMCOLS )
	{
	    x = x + max_width + COLSPACE;
	}
	else
	{
	    x = COLSPACE;
	    y = y + max_ascent + max_descent + ROWSPACE;
	    column = 0;
	}

	printf( "menu[%d] - %s, top left (%d,%d), bot right (%d,%d)\n",
	       i, menu[i].test_info->name,
	       menu[i].top_left_x, menu[i].top_left_y,
	       menu[i].bot_right_x, menu[i].bot_right_y );
    }
    width = (( NUMCOLS + 1 ) * COLSPACE ) + ( NUMCOLS * max_width );
    if( column != 0 )
    {
	height = y + max_ascent + max_descent + ROWSPACE;
    }
    else
    {
	height = y;
    }
/*
 * create the top window and size accordingly
 */
    fprintf( stderr, "creating top window\n" );
    black_pix = BlackPixel( dpy, DefaultScreen( dpy ));
    white_pix = WhitePixel( dpy, DefaultScreen( dpy ));
    top_win = XCreateSimpleWindow( dpy, DefaultRootWindow( dpy ), 60, 60,
				  width, height, 4, black_pix, white_pix );
    size_hints.flags = PPosition | PSize | PMinSize;
    size_hints.x = 60;
    size_hints.y = 60;
    size_hints.width = width;
    size_hints.height = height;
    size_hints.min_width = width;
    size_hints.min_height = height;

    XSetStandardProperties( dpy, top_win, WINNAME, ICONNAME, None, NULL, 0,
			   &size_hints );
/*
 * create the delete atom
 */
    del_atom = XInternAtom( dpy, DELATOMNAME, FALSE );
    XChangeProperty( dpy, top_win, del_atom, XA_INTEGER, 8, PropModeAppend,
		    NULL, 0 );
/*
 * create the default GCs
 */
    fprintf( stderr, "Creating GCs\n" );
    gc_vals.function = GXcopy;
    gc_vals.foreground = black_pix;
    gc_vals.background = white_pix;
    gc_vals.font = font_info->fid;

    normGC = XCreateGC( dpy, top_win,
		       ( GCFunction | GCForeground | GCBackground | GCFont ),
		       &gc_vals );
    gc_vals.foreground = white_pix;
    gc_vals.background = black_pix;
    revGC = XCreateGC( dpy, top_win,
		      ( GCFunction | GCForeground | GCBackground | GCFont ),
		      &gc_vals );
/*
 * initialize the global data
 */
    fprintf( stderr, "mallocing defaults and run_data\n" );
    defaults = (Defaults)malloc( sizeof( *defaults ));
    defaults->dpy = dpy;
    defaults->top_win = top_win;
    defaults->normGC = normGC;
    defaults->revGC = revGC;
    defaults->font_info = font_info;
    defaults->shm_flag = FALSE;
    defaults->black_pix = black_pix;
    defaults->white_pix = white_pix;
    defaults->del_atom = del_atom;

    run_data = (struct _run_data *)malloc( sizeof( *run_data ));
    run_data->defaults = defaults;
    run_data->menus = menu;
    run_data->num_items = num_items;
    run_data->children = NULL;
    run_data->context = context;
    fprintf( stderr, "top context = %lx\n", context );
    fprintf( stderr, "top window = %ld\n", top_win );
/*
 * register the handlers
 */
    click_handler = TKAddEventHandler( context, ClickHandler, run_data,
					    top_win, ButtonPressMask );
    expose_handler = TKAddEventHandler( context, ExposeHandler, run_data,
					     top_win, ExposureMask );
    fprintf( stderr, "click handler = %lx\n", click_handler );
    fprintf( stderr, "expose handler = %lx\n", expose_handler );
/*
 * map the window and call MainLoop
 */
    fprintf( stderr, "Mapping window and calling main loop\n" );
    defaults->state = TEST_RUNNING;
    XMapWindow( dpy, top_win );
    TKMainLoop( context );
/*
 * cleanup and exit
 */
    fprintf( stderr, "Calling cleanup\n" );
    Cleanup( run_data );
    exit( EXIT_SUCCESS );

}	/* end of RunTest function */

/*
 * start of Cleanup function
 * shutdown each of the children and frees up the resources
 */

static void
Cleanup( run_data )
    struct _run_data	*run_data;
{
    Display	*dpy = run_data->defaults->dpy;
    struct _child	*child;
    struct _child	*next;
    pthread_addr_t	status;
/*
 * close the window and notify the children
 */
    XUnmapWindow( dpy, run_data->defaults->top_win );
    XDeleteProperty( dpy, run_data->defaults->top_win,
		    run_data->defaults->del_atom );
    XFlush( dpy );
/*
 * collect all the children
 */
    child = run_data->children;
    while( child != NULL )
    {
	pthread_join( child->thread, &status );
	next = child->next;
	free( child );
	child = next;
    }
/*
 * sync up, kill the context, get rid of the window and shut the display
 */
    XSync( dpy, TRUE );
    TKDestroyAppContext( run_data->context );
    XDestroyWindow( dpy, run_data->defaults->top_win );
    XCloseDisplay( dpy );

    return;

}	/* end of Cleanup function */

/*
 * start of FindItem function
 * returns the menu item that was selected
 */

static int
FindItem( menu, num_items, x, y )
    MenuList	menu;
    int		num_items;
    long int	x;
    long int	y;
{
    int		i;

    for( i = 0; i < num_items; i++ )
    {
	if(( x >= menu[i].top_left_x ) && ( x <= menu[i].bot_right_x )
	   && ( y >= menu[i].top_left_y ) && ( y <= menu[i].bot_right_y ))
	{
	    return( i );
	}
    }
    return( -1 );

}	/* end of FindItem function */

/*
 * start of DrawItem function
 * draws one menu item in the window
 */

static void
DrawItem( dpy, win, fg, bg, menu )
    Display	*dpy;
    Window	win;
    GC		fg;
    GC		bg;
    MenuList	menu;
{
    long int	x;
    long int	y;
    long int	width;
    long int	height;

    x = menu->top_left_x - BORDER;
    y = menu->top_left_y - BORDER;
    width = menu->bot_right_x - menu->top_left_x + 2 * BORDER;
    height = menu->bot_right_y - menu->top_left_y + 2 * BORDER;
    XFillRectangle( dpy, win, bg, x, y, width, height );
    XDrawRectangle( dpy, win, fg, x, y, width, height );
    XDrawString( dpy, win, fg, menu->bot_left_x, menu->bot_left_y,
		menu->test_info->name, strlen( menu->test_info->name ));
    
}	/* end of DrawItem function */

/*
 * start of DrawMenu function
 * draws each menu item in the window
 * (could be optimized to do partial redraws)
 */

static void
DrawMenu( dpy, win, fg, bg, menu, num_items )
    Display	*dpy;
    Window	win;
    GC		fg;
    GC		bg;
    MenuList	menu;
    int		num_items;
{
    int		i;

    for( i = 0; i < num_items; i++ )
    {
	DrawItem( dpy, win, fg, bg, &menu[i] );
    }
}	/* end of DrawMenu function */

/*
 * start of ClickHandler function
 * handles mouse button events for the top window
 */

static void
ClickHandler( context, window, event, data )
    AppContext	context;
    Window	window;
    XEvent	*event;
    void	*data;
{
    Defaults		defaults = ((struct _run_data*)data)->defaults;
    struct _child	*child;
    MenuList		item;
    int			item_num;
/*
 * find which menu item was selected
 */
    item_num = FindItem( ((struct _run_data*)data)->menus,
		    ((struct _run_data*)data)->num_items,
		    event->xbutton.x, event->xbutton.y );
    fprintf( stderr, "ClickHandler: x = %d, y = %d, item = %d\n", event->xbutton.x, event->xbutton.y, item_num );
    if( item_num >= 0 )
    {
	/* highlight (rev-video) the selected item */
	item = &((struct _run_data*)data)->menus[item_num];
	DrawItem( defaults->dpy, defaults->top_win, defaults->revGC,
		 defaults->normGC, item );
	XFlush( defaults->dpy );
	/* fork a new thread to run the test */
	if( item->test_info->proc == NULL )
	{
	    TKExitMainLoop( context );
	}
	else
	{
	    item->test_info->defaults = defaults;
	    child = (struct _child*)malloc( sizeof( *child ));
	    pthread_create( &child->thread, pthread_attr_default,
			   (void*(*)())ChildProc, item->test_info );
	    child->next = ((struct _run_data*)data)->children;
	    ((struct _run_data*)data)->children = child;
	}
	/* un-highlight the selected item */
	DrawItem( defaults->dpy, defaults->top_win, defaults->normGC,
		 defaults->revGC, item );
	XFlush( defaults->dpy );
    }

}	/* end of ClickHandler function */

/*
 * start of ExposeHandler function
 * handles expose events for the top window - simply redraw the menu
 * could be optimized to only redraw the exposed parts
 */

static void
ExposeHandler( context, window, event, data )
    AppContext	context;
    Window	window;
    XEvent	*event;
    void	*data;
{
    Defaults	defaults = ((struct _run_data*)data)->defaults;

    fprintf( stderr, "ExposeHandler: window = %d\n", event->xany.window );
    DrawMenu( defaults->dpy, defaults->top_win, defaults->normGC,
	     defaults->revGC, ((struct _run_data*)data)->menus,
	     ((struct _run_data*)data)->num_items  );

}	/* end of ExposeHandler function */

/*
 * start of ChildProc function
 * initializes the test environment and calls the test proc
 */

static void
ChildProc( test_info )
    TestInfo	test_info;
{
    Defaults	defaults = test_info->defaults;
    AppContext	context;
    TestData	test_data;
    XSizeHints	size_hints;
    HandlerList	del_prop_handler;
    HandlerList ctl_btn1_handler;
/*
 * create and initialize the test data structure
 */
    test_data = (TestData)malloc( sizeof( *test_data ));
    test_data->defaults = defaults;
    test_data->data = test_info->data;
/*
 * create a window for the test and set the props
 */
    test_info->geometry.width = MAX( test_info->geometry.width, 50 );
    test_info->geometry.height = MAX( test_info->geometry.height, 50 );
    test_data->window = XCreateSimpleWindow( defaults->dpy,
					    DefaultRootWindow( defaults->dpy ),
					    test_info->geometry.x,
					    test_info->geometry.y,
					    test_info->geometry.width,
					    test_info->geometry.height,
					    2, defaults->black_pix,
					    defaults->white_pix );
    size_hints.flags = PPosition | PSize | PMinSize;
    size_hints.x = test_info->geometry.x;
    size_hints.y = test_info->geometry.y;
    size_hints.width = test_info->geometry.width;
    size_hints.height = test_info->geometry.height;
    size_hints.min_width = test_info->geometry.width;
    size_hints.min_height = test_info->geometry.height;
    XSetStandardProperties( defaults->dpy, test_data->window, test_info->name,
			   test_info->name, None, NULL , 0, &size_hints );
/*
 * create a new context for the test and add a couple of default handlers
 */
    context = TKCreateAppContext( defaults->dpy );
    ctl_btn1_handler = TKAddEventHandler( context, CtlBtn1Handler, NULL,
					 test_data->window, ButtonPressMask );
    del_prop_handler = TKAddEventHandler( context, DelPropHandler, defaults,
					 defaults->top_win,
					 PropertyChangeMask );
    fprintf( stderr, "child context = %lx\n", context );
    fprintf( stderr, "child window = %lx\n", test_data->window );
    fprintf( stderr, "prop window = %lx\n", defaults->top_win );
    fprintf( stderr, "ctrl button handler = %lx\n", ctl_btn1_handler );
    fprintf( stderr, "delete prop handler = %lx\n", del_prop_handler );
/*
 * call the test initialization procedure
 */
    test_info->proc( context, test_data );
/*
 * map the window and call main loop
 */
    XMapWindow( defaults->dpy, test_data->window );
    TKMainLoop( context );
/*
 * clean up and exit
 */
    TKRemoveEventHandler( context, ctl_btn1_handler );
    TKRemoveEventHandler( context, del_prop_handler );
    XDestroyWindow( defaults->dpy, test_data->window );
    TKDestroyAppContext( context );

}	/* end of ChildProc function */

/*
 * start of CtlBtn1Handler function
 * handles ctrl-btn1 event to quit a test
 */

static void
CtlBtn1Handler( context, window, event, data )
    AppContext	context;
    Window	window;
    XEvent	*event;
    void	*data;
{
    fprintf( stderr, "CtlBtn1Handler: window = %d\n", event->xany.window );
    if((( event->xbutton.state & ControlMask ) == ControlMask ) && ( event->xbutton.button == Button1 ))
    {
	TKExitMainLoop( context );
    }

}	/* end of CtlBtn1Handler function */

/*
 * start of DelPropHandler function
 * handles ctrl-btn1 event to quit a test
 */

static void
DelPropHandler( context, window, event, data )
    AppContext	context;
    Window	window;
    XEvent	*event;
    void	*data;
{
    fprintf( stderr, "DelPropHandler: window = %d\n", event->xany.window );
    if(( event->xproperty.window == ((Defaults)data)->top_win )
       && ( event->xproperty.state == PropertyDelete )
       && ( event->xproperty.atom == ((Defaults)data)->del_atom ))
    {
	TKExitMainLoop( context );
    }

}	/* end of DelPropHandler function */

