/*
 * XlibTK.h - header file for toolkit to test the MT Xlib
 */

#ifndef _XLIBTK_H_
#define _XLIBTK_H_	1

#include	<X11/Xlib.h>

/*
 * application context states
 */

#define	AC_STOPPED	0
#define	AC_RUNNING	1
#define AC_DISPATCH	2
#define	AC_SHUTDOWN	3

typedef	int		Boolean;

/*
 * useful stuff
 */

#ifndef MAX
#define	MAX(a,b)	(((a) > (b)) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#endif

/*
 * HandlerList structure
 */

typedef struct _handler_list
{
    struct _handler_list	*next;	/* next handler in list */
    struct _handler_list	*prev;	/* next handler in list */
    struct _window_list	*window;	/* window this handler acts on */
    unsigned long	mask;		/* event mask for this handler */
    Boolean		non_maskable;	/* are non_maskable events handled? */
    Boolean		valid;		/* has this handler been removed? */
    void		(*handler)();	/* actual handler proc */
    void		*data;		/* data to be sent to handler */
    void		*timeout;	/* time to call this handler */
}	HandlerStruct, *HandlerList;

/*
 * WindowList structure
 */

typedef struct _window_list
{
    struct _window_list	*next;		/* next window in list */
    Window		window;		/* window ID (from server) */
    unsigned long	mask;		/* event mask for this window */
    HandlerList		handlers;	/* event handlers for this window */
}	*WindowList;

/*
 * AppContext structure
 */

typedef struct _app_context
{
    struct _app_context	*next;		/* next application context in list */
    Display		*dpy;		/* display connection */
    XQueue		*queue;		/* event queue for context */
    XQueueProp		props;		/* event queue properties */
    HandlerList		event_procs;	/* event handlers */
    HandlerList		cleanup_procs;	/* cleanup procs */
    HandlerList		timeout_procs;	/* timeout procs */
    HandlerList		add_list;	/* handlers that need to be added */
    Boolean		remove_flag;	/* have handlers been removed? */
    WindowList		windows;	/* windows in context */
    pthread_t		thread;		/* thread currently using context */
    pthread_mutex_t	lock;		/* lock for context data */
    pthread_mutex_t	event_lock;	/* lock for event handler list */
    pthread_mutex_t	cleanup_lock;	/* lock for cleanup handler list */
    pthread_mutex_t	timeout_lock;	/* lock for timeout handler list */
    pthread_mutex_t	window_lock;	/* lock for window list */
    char		status;		/* current status of context */
}	*AppContext;

/*
 * geometry structure
 */

typedef struct _geometry
{
    int		x;
    int		y;
    unsigned	width;
    unsigned	height;
}	Geometry;

AppContext TKInitialize(
#if NeedFunctionPrototypes
    char *	/* dpy_name */
#endif
);
Display *TKDpyFromContext(
#if NeedFunctionPrototypes
    AppContext	/* context */
#endif
);
AppContext TKCreateAppContext(
#if NeedFunctionPrototypes
    Display *	/* dpy */
#endif
);
Boolean TKDestroyAppContext(
#if NeedFunctionPrototypes
    AppContext	/* context */
#endif
);
HandlerList TKAddEventHandler(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    void (*)()	/* handler */,
    void *	/* data */,
    Window	/* window */,
    unsigned long	/* mask */
#endif
);

#endif	/* ! _XLIBTK_H_ */
